/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "ssft.h"
#include "assert.h"

using namespace std;

ClientConnection::ClientConnection(SOCKET conn, SSL_CTX *mpCtx) : 
mSocket(conn),
mState(STATE_CREATED),
mpCtx(mpCtx),
mpSsl(NULL)
{
}

ClientConnection::~ClientConnection()
{
   Close();
}

void ClientConnection::Close()
{
    if (NULL != mpSsl)
    {
        if (debug_flag)
            log << LOG_DATE << "Closing SSL connection." << ENDL;
        
        SSL_shutdown(mpSsl); 
        SSL_free(mpSsl);
        mpSsl = NULL;
    }
    
    if (-1 != mSocket)
    {
        closesocket(mSocket);
        mSocket = -1;
    }
}

bool ClientConnection::Create() throw(SsftDisconnect)
{
    int err;

    if ((mpSsl = SSL_new(mpCtx)) == NULL)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_new() failed " 
            << ssl_err << ENDL;

        return false;
    }

    if (SSL_set_fd(mpSsl, mSocket) == 0)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_set_fd() failed " 
            << ssl_err << ENDL;

        return false;
    }

    if (debug_flag)
        log << "Doing SSL_connect()" << ENDL;
    
    if ((err = SSL_connect(mpSsl)) == 0)
    {
        SslError ssl_err(mpSsl, err);

        log << LOG_DATE  
            << "ERROR - SSL_connect() failed with result: " 
            << err
            << ". SSL error: "
            << ssl_err 
            << ENDL;

        return false;
    }

    
    X509 *server_cert = NULL;

    server_cert = SSL_get_peer_certificate (mpSsl);
    if (!server_cert)
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - Failed to retrieve certificate from server "
            << ssl_err << ENDL;

        X509_free(server_cert);
        return false;
    }

    if (verbose_flag)
    {
        log << LOG_DATE 
            << "SSL connection using "
            << SSL_get_cipher(mpSsl)
            << ENDL;
        
        char buf[1024 * 4];
        *buf = 0;
        X509_NAME_oneline (X509_get_subject_name (server_cert), buf, sizeof(buf));

        log << LOG_DATE
            << "Server certificate: "
            << buf
            << ENDL;

        X509_NAME_oneline(X509_get_issuer_name (server_cert),buf , sizeof(buf));

        log << LOG_DATE
            << "Certificate issuer: "
            << buf
            << ENDL;
        
        if ((err = SSL_get_verify_result(mpSsl)) != X509_V_OK)
        {
            SslError ssl_err(mpSsl, err);
            
            log << LOG_DATE  
                << "ERROR - Server certificate verification failed. Status= "
                << err
                << " SSL error: "
                << ssl_err << ENDL;
            
            X509_free(server_cert);
            return false;
        }
    }

    X509_free(server_cert);
   
    SetState(STATE_CONNECTED);

    // Do initial handshake with server
    // Get initial tag + version

    if (debug_flag)
        log << LOG_DATE  << "Getting server tag and version" << ENDL;

    size_t got_bytes = 0;
#define INITIAL_HDR_LEN 24
    char hdr_buf[INITIAL_HDR_LEN];
    memset(hdr_buf, 0, INITIAL_HDR_LEN);
    while(got_bytes < INITIAL_HDR_LEN)
    {
        int bytes = SSL_read(mpSsl, 
            hdr_buf + got_bytes,  
            INITIAL_HDR_LEN - got_bytes);

        if (bytes > 0)
        {
            got_bytes += bytes;
        }
        else if (0 == bytes)
        {
            log << LOG_DATE  
                << "ERROR - Unexcpected end of transmission during initial handshake"
                << ENDL;
            
            return false;
        }
        else 
        {
            // Error
            SslError ssl_err;
            
            log << LOG_DATE  
                << "ERROR - Network error during initial handshake: :" 
                << ssl_err 
                << " Socket error: " 
                << GET_SOCKET_ERROR
                << ENDL;
            
            return false;
        }
    }

    // Got header
    if (strncmp(SSFT_SVR_HELO, hdr_buf, strlen(SSFT_SVR_HELO)) != 0)
    {
        log << LOG_DATE  
            << "ERROR - Bad helo from server" 
            << err << ENDL;

        return false;
    }
    
    mServerVersion = (SsftProtocolVersionsE) ntohl(*(unsigned int *)
        (hdr_buf + strlen(SSFT_SVR_HELO)));

    switch(mServerVersion)
    {
    default:
        // Fall trough to the highest protocol we support
    case SSFT_VERSION_1:
        // Send header
        {
            if (debug_flag)
                log << LOG_DATE  << "Using SSFT_VERSION_1 - sending header" << ENDL;

            char hdr[8];
            hdr[0] = 'S';
            hdr[1] = 'S';
            hdr[2] = 'F';
            hdr[3] = 'T';
            *(unsigned int *)(&hdr[4]) = htonl(SSFT_VERSION_1);
            
            if (SSL_write(mpSsl, hdr, 8) < 8)
            {
                SslError ssl_err;
                
                log << LOG_DATE  
                    << "ERROR - Network error during initial handshake: :" 
                    << ssl_err << ENDL;
                
                return false;
            }
        }
    }

    SsftResultCodesE result = GetServerReply();

    if (debug_flag)
    {
        log << LOG_DATE
            << "Server replied "
            << result
            << ENDL;
    }

    return SSFT_RESULT_OPERATION_OK == result;
}

const char *ClientConnection::GetStateName() const
{
    switch(mState)
    {
    case STATE_CREATED:
        return "STATE_CREATED";
    case STATE_CONNECTED:
        return "STATE_CONNECTED";
    case STATE_SENT_FIRST_HEADER:
        return "STATE_SENT_FIRST_HEADER";
    case STATE_SENT_CMD_HEADER:
        return "STATE_SENT_CMD_HEADER";
    case STATE_SENDING:
        return "STATE_SENDING";
    default:
        return "**Invalid state value **";
    }
}

std::string ClientConnection::Explain() const
{
    return "server";
}

void ClientConnection::SetState(StateE newState)
{
    mState = newState;
}

SsftResultCodesE ClientConnection::DeletePath(const char *filename) 
throw(SsftDisconnect)
{
    return GenericCmd(SSFT_CMD_DELETE_FILE_OR_DIR, filename);
}

SsftResultCodesE ClientConnection::CreateDir(const char *filename) 
throw(SsftDisconnect)
{
    return GenericCmd(SSFT_CMD_CREATE_DIR, filename);
}

SsftResultCodesE ClientConnection::GenericCmd(const SsftCmdE cmd,
                                              const char *filepath) 
                                              throw(SsftDisconnect)
{
    XmitBuffer hdr;

    int flags = 0;

    if (recurse_flag)
        flags |= DIRFLAG_RECURSE;

    hdr << cmd
        << filepath
        << flags // flags
        << hash_digest;

    if (debug_flag)
    {
        log << LOG_DATE
            << "Excecuting the command "
            << cmd
            << " on filepath \""
            << filepath
            << "\"."
            << ENDL;
    }

    Send(hdr.GetSendVector());
    SetState(STATE_SENT_CMD_HEADER);

    SsftResultCodesE result = GetServerReply();
    if (result != SSFT_RESULT_OPERATION_OK)
    {
        log << LOG_DATE
            << "The command: "
            << cmd
            << " with filepath \""
            << filepath
            << "\" failed. The server replied "
            << result
            << ENDL;
    }

    return result;
}

void ClientConnection::ListDir(const char *filename, DirHandler& outHandler) 
 throw(SsftDisconnect, TransferException)
{
    XmitBuffer hdr;

    int flags = 0;

    if (recurse_flag)
        flags |= DIRFLAG_RECURSE;

    hdr << SSFT_CMD_DIR_SEVER_TO_CLI
        << filename
        << flags // flags
        << hash_digest;

    Send(hdr.GetSendVector());
    SetState(STATE_SENT_CMD_HEADER);

    SsftResultCodesE result = GetServerReply();
    if (result != SSFT_RESULT_OPERATION_OK)
    {
        log << LOG_DATE
            << "Failed to list directory \""
            << filename
            << "\". The server replied "
            << result
            << ENDL;

        throw TransferException();
    }

    // Receive
    bool have_all = false;
    while(!have_all)
    {
        // Get a number of complete dir-nodes
        GetServerMessage();

        while(!have_all && !mReplyHdr.IsAtEndOfOutput())
        {
            DirList::DirTagsE current_tag;
            mReplyHdr >> (int &)current_tag;

            switch(current_tag)
            {
            case DirList::TAG_DIR_HDR: // New directory
                {
                    std::string dir_name;
                    
                    mReplyHdr >> dir_name;
                    
                    outHandler.OnDirEntry(dir_name);
                }
                break;
            case DirList::TAG_NODE: // Entry
                {
                    std::string name, hash_val;
                    DirNodeTypesE node_type;
                    ssft_flen_t len;
                    ssft_date_t mtime;

                    mReplyHdr >> name;
                    mReplyHdr >> (int &)node_type;
                    mReplyHdr >> len;
                    mReplyHdr >> mtime;
                    mReplyHdr >> hash_val;

                    outHandler.OnNodeEntry(name, node_type, len, mtime, 
                        hash_digest.empty() ? NULL : hash_val.c_str());
                }
                break;
            case DirList::TAG_END: // Done
                have_all = true;
                break;
            }
        }
    }
}

void ClientConnection::RecvFile(const char *filename, 
                                const std::string& dst_path) 
                                throw(SsftDisconnect, TransferException)
{
    string my_path = dst_path;

    if (is_dir(my_path))
    {
        const char *p = filename, *name = p;

        while(*p)
        {
            if (IS_SLASH(*p))
                name = p + 1;
            ++p;
        }

        if (*name)
            my_path = merge_paths(my_path, string(name));
    }

    if (is_dir(my_path))
    {
        log << LOG_DATE  << "Error: file \""
            << filename
            << "\" is a directory!" 
            << dst_path
            << ENDL;
        
        throw TransferException("");
    }

    // Check that we can reach the file
    if (recurse_flag)
        make_dirs(my_path, 0, NULL, NULL);

    int fd = ssft_open(my_path.c_str(),  
        O_CREAT | O_SEQUENTIAL | O_TRUNC | O_WRONLY | O_BINARY,
        S_IREAD | S_IWRITE);
    if (fd < 0)
    {
        log << LOG_DATE  
            << "Failed to open file \"" 
            << my_path 
            << "\" for write!" << ENDL;

        throw TransferException("");
    }

    XmitBuffer hdr;

    hdr << SSFT_CMD_TRANSF_SEVER_TO_CLI
        << filename;

    Send(hdr.GetSendVector());
    SetState(STATE_SENT_CMD_HEADER);

    SsftResultCodesE result = GetServerReply();
    if (result != SSFT_RESULT_OPERATION_OK)
    {
        log << LOG_DATE
            << "Failed to receive file \""
            << my_path
            << "\". The server replied "
            << result
            << ENDL;

        close(fd);
        throw TransferException("");
    }

    // Extract file details
    ssft_flen_t file_len = 0, bytes_left = 0;
    ssft_date_t file_date = 0;
    mReplyHdr >> file_len;
    mReplyHdr >> file_date;

    log << LOG_DATE  << "Receiving file \""
        << my_path
        << "\". "
        << file_len
        << " bytes."
        << ENDL;

    bytes_left = file_len;
    while(bytes_left)
    {
        char buffer[SSFT_NET_BUFLEN];

        size_t packet_size = min(sizeof(buffer), bytes_left);

        int bytes = SSL_read(mpSsl, buffer, packet_size);

        if (bytes == 0)
        {
            SslError ssl_err;
            int sck_err = GET_SOCKET_ERROR;
            
            log << LOG_DATE  
                << "ERROR - The server unexcpectedly closed the connection :" 
                << ssl_err 
                << " Socket error: " << sck_err
                << ENDL;
            
            close(fd);
            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
        }
        else if (bytes < 0)
        {
            // Error
            SslError ssl_err;
            int sck_err = GET_SOCKET_ERROR;
            
            log << LOG_DATE  
                << "ERROR - Network error during read: :" 
                << ssl_err 
                << " Socket error: " << sck_err
                << ENDL;

            close(fd);
            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
        }

        assert(bytes > 0);

        if (write(fd, buffer, bytes) != bytes)
        {
            log << LOG_DATE  
                << "Write error on file: "
                << strerror(errno)
                << ENDL;

            close(fd);
            throw SsftDisconnect(SSFT_RESULT_READ_ERROR);
        }

        bytes_left -= bytes;
    }

    close(fd);

    log << LOG_DATE << "File transfered OK" << ENDL;
}

void ClientConnection::SendFile(const char *filename, 
                                const std::string& dst_path) 
                                throw(SsftDisconnect, TransferException)
{
    log << LOG_DATE  << "Sending file \""
        << filename
        << "\" --> \"" 
        << hostname
        << ':'
        << dst_path
        << "\""
        << ENDL;
    
    int fd = ssft_open(filename, O_RDONLY | O_BINARY);
    if (fd < 0)
    {
        log << LOG_DATE  
            << "Failed to open file \"" 
            << filename 
            << "\" for read!" << ENDL;

        throw TransferException();
    }

    ssft_stat_t st;
    int result = ssft_fstat(fd, &st);

    if (0 != result)
    {
        log << LOG_DATE 
            << "stat(\""
            << filename
            << "\") failed "
            << errno
            << " "
            << strerror(errno)
            << ENDL;
        
        close(fd);
        throw TransferException();
    }

    if ((st.st_mode & S_IFREG) != S_IFREG)
    {
        log << LOG_DATE  
            << "Not a regulare file \"" 
            << filename 
            << "\"" << ENDL;
        close(fd);
        throw TransferException();
    }
    
    XmitBuffer hdr;

    hdr << SSFT_CMD_TRANSF_CLI_TO_SEVER
        << filename
        << dst_path
        << (ssft_flen_t)st.st_ctime * (ssft_flen_t)1000
        << (ssft_flen_t)st.st_size;

    Send(hdr.GetSendVector());
    SetState(STATE_SENT_CMD_HEADER);
    
    // Get server response
    SsftResultCodesE server_reply = GetServerReply();

    if (server_reply != SSFT_RESULT_OPERATION_OK)
    {
        close(fd);
        throw TransferException();
    }

    // Send data
    ssft_flen_t bytes_left = st.st_size, bytes_sent = 0;
    while(bytes_left)
    {
        char buffer[SSFT_NET_BUFLEN];

        size_t packet_size = min(sizeof(buffer), bytes_left);
        int bytes = read(fd, buffer, packet_size);
        
        if (0 >= bytes)
        {
            log << LOG_DATE  
                << "Read error from file"
                << ENDL;
            close(fd);
            throw SsftDisconnect(SSFT_RESULT_READ_ERROR);
        }

        bytes_left -= bytes;

        if (SSL_write(mpSsl, buffer, bytes) != bytes)
        {
            // Error
            SslError ssl_err;
            int sck_err = GET_SOCKET_ERROR;
            
            log << LOG_DATE  
                << "ERROR - Network error during write: :" 
                << ssl_err 
                << " Socket error: " << sck_err
                << ENDL;

            close(fd);
            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
        }

        bytes_sent += bytes;
    }

    if (debug_flag)
    {
        log << LOG_DATE 
            << "Transfer complete. Sent "
            << bytes_sent
            << " bytes."
            << ENDL;
    }

    close(fd);

    // Get server resposne
    server_reply = GetServerReply();

    if (SSFT_RESULT_OPERATION_OK == server_reply)
    {
        log << LOG_DATE << "File transfered OK" << ENDL;
    }
    else
    {
        log << LOG_DATE <<
            "File transfer failed with error "
            << server_reply
            << ENDL;

        throw TransferException();
    }
}

// Returns true if more data can be sent
bool ClientConnection::Send(const std::vector<char>& sndBuf) 
throw(SsftDisconnect)
{
    size_t orig_len = mOutBuf.size();
    mOutBuf.resize(orig_len + sndBuf.size());
    memcpy(&mOutBuf[orig_len], &sndBuf[0], sndBuf.size());

    if (debug_flag)
    {
        log << LOG_DATE 
            << "Sending "
            << sndBuf.size()
            << " bytes."
            << ENDL;
    }

    return Send();
}

bool ClientConnection::Send() throw(SsftDisconnect)
{
again:
    int bytes = SSL_write(mpSsl, &mOutBuf[0], mOutBuf.size());
    
    if (debug_flag)
    {
        log << LOG_DATE 
            << "Wrote "
            << bytes
            << " bytes."
            << ENDL;
    }
    
    int status = SSL_get_error(mpSsl, bytes);
    switch(status)
    {
    case SSL_ERROR_NONE:
        if (bytes != (int)mOutBuf.size())
        {
            memmove(&mOutBuf[0], &mOutBuf[bytes],  mOutBuf.size() - bytes);
            mOutBuf.resize(mOutBuf.size() - bytes);
            goto again;
        }
        else
        {
            mOutBuf.resize(0);
        }
        break;
    case SSL_ERROR_WANT_WRITE:
        return false;
    default:
        {
            // Error
            SslError ssl_err;
            int sck_err = GET_SOCKET_ERROR;
            
            log << LOG_DATE  
                << "ERROR - Network error during initial handshake: :" 
                << ssl_err 
                << " Socket error: " << sck_err
                << ENDL;
            
            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
        }
        break;
    }

    return true;
}

SsftResultCodesE ClientConnection::GetServerReply() throw(SsftDisconnect)
{
    GetServerMessage();

    // decode header
    int reply_code = 0;
    mReplyHdr >> reply_code;

    return (SsftResultCodesE)reply_code;
}

void ClientConnection::GetServerMessage() throw(SsftDisconnect)
{
    mReplyHdr.Reset();

    int got_bytes = 0;
    int want_bytes = 4; // Header length

    while(got_bytes < want_bytes)
    {
        int bytes = SSL_read(mpSsl, 
            mReplyHdr.GetRecvBuffer(want_bytes) + got_bytes, 
            want_bytes - got_bytes);

        int status = SSL_get_error(mpSsl, bytes);
        switch(status)
        {
        case SSL_ERROR_NONE:
            got_bytes += bytes;
            break;
        case SSL_ERROR_ZERO_RETURN:
            // EOF
            log << LOG_DATE  
                << "The network connection was unexpectedly closed"
                << ENDL;
            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
            break;
        case SSL_ERROR_WANT_READ:
            throw SsftWouldBlock();
            break;
        default:
            {
                // Error
                SslError ssl_err;
                
                log << LOG_DATE  
                    << "ERROR - Network error during read: " 
                    << ssl_err 
                    << " Socket error: " 
                    << GET_SOCKET_ERROR
                    << ENDL;
                
                throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
            }
            break;
        } // switch( status)

        if ((4 == got_bytes) && (4 == want_bytes))
        {
            // Get header length
            mReplyHdr >> want_bytes;

            if (want_bytes > XmitBuffer::MAX_BUF_LEN)
            {
                // Invalid reply.
                log << LOG_DATE  
                    << "ERROR - Received an invalid (overflow) header "
                    "length from the server: "
                    << want_bytes
                    << ENDL;
                throw SsftDisconnect(SSFT_RESULT_INVALID_REPLY);
            }

            if (debug_flag)
            {
                log << LOG_DATE
                    << "Server reply header length: "
                    << want_bytes
                    << ENDL;
            }

            if (4 >= want_bytes)
            {
                // Invalid reply.
                log << LOG_DATE  
                    << "ERROR - Received an invalid (underflow) reply header "
                    "length from the server: "
                    << want_bytes
                    << ENDL;
                throw SsftDisconnect(SSFT_RESULT_INVALID_REPLY);
            }
        }
    }
}

void ClientConnection::Quit()
{
    if (debug_flag)
        log << LOG_DATE << "Sending QUIT command." << ENDL;

    
    XmitBuffer hdr;
    hdr << SSFT_CMD_QUIT;

    Send(hdr.GetSendVector());
    GetServerReply();
}

