/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#ifndef DIR_HANDLER_H
#define DIR_HANDLER_H

class DirHandler
{
public:
    DirHandler() {}
    virtual ~DirHandler() {}

    virtual void OnDirEntry(const std::string& dirName) = 0;
    virtual void OnNodeEntry(const std::string nodeName,
        DirNodeTypesE nodeType,
        ssft_flen_t nodeLength,
        ssft_date_t nodeTime,
        const char *hashValue) = 0;
    virtual void OnDone() = 0;
};

template <class outT>
class DirHandlerOut : public DirHandler
{
public:
    DirHandlerOut(outT& out)
        : mOut(out)
    {
    }

    virtual void OnDirEntry(const std::string& dirName)
    {
        mOut << dirName << ":" << std::endl;
    }

    virtual void OnNodeEntry(const std::string nodeName,
        DirNodeTypesE nodeType,
        ssft_flen_t nodeLength,
        ssft_date_t nodeTime,
        const char *hashValue)
    {
        char buf[64], tbuf[32];

        time_t mtime = (nodeTime / 1000);
        struct tm *tm = localtime(&mtime);
        if (tm)
            strftime(tbuf, sizeof(tbuf), date_fmt.c_str(), tm);
        else
            strcpy(tbuf, "****");

        switch(nodeType)
        {
        case TYPE_FILE:
            sprintf(buf, 
#ifdef WIN32
		    "%12I64d %20s ",
#else
		    "%12lld %20s ",
#endif
                nodeLength,
                tbuf);
            break;
        case TYPE_DIR:
            sprintf(buf, "%12s %20s ",
                "<dir>",
                tbuf);
            break;
        default:
            return; // Unsupported type
        }
        
        if (hashValue)
            mOut << buf << hashValue << ' ' << nodeName << std::endl;
        else
            mOut << buf << nodeName << std::endl;
    }

    virtual void OnDone()
    {
    }

private:
    outT& mOut;
};


class DirHandlerList : public DirHandler
{
public:
    DirHandlerList(fileset_t& dstList, 
		   const std::string& dstPath)
        : mList(dstList), mDstPath(dstPath)
    {
	 mDstPath.c_str();
	 char *p = &mDstPath[0];
	 for(; *p; p++)
	 {
	      if (IS_SLASH(*p))
		   *p = '/';
	 }

	 if (mDstPath.size() > 1)
	 {
	      if (mDstPath[mDstPath.size() -1] == '/')
		   mDstPath.erase(mDstPath.size() -2);
	 }
    }

    virtual void OnDirEntry(const std::string& dirName)
    {
	 if (dirName.size() >= mDstPath.size())
	      mCurrentPath = dirName.c_str() + mDstPath.size();
	 else
	      mCurrentPath = dirName;

	 if (mCurrentPath.empty())
	      mCurrentPath = "/";
    }

    virtual void OnNodeEntry(const std::string nodeName,
        DirNodeTypesE nodeType,
        ssft_flen_t nodeLength,
        ssft_date_t nodeTime,
        const char *hashValue)
    {
        
        FileInfo node;

        const char *name = nodeName.c_str();
        if (!strcmp(name, ".") || !strcmp(name, ".."))
            return;
        
        node.mName = merge_paths(mCurrentPath, 
            nodeName, '/');
        node.mType = nodeType;
        node.mLength = nodeLength;
        node.mDate = nodeTime;
        if (hashValue)
            node.mHash = hashValue;
        
        mList.insert(node);
    }
    
    virtual void OnDone()
    {
    }

    fileset_t& mList;
    std::string mDstPath;
    std::string mCurrentPath;
};

#endif // DIR_HANDLER_H
