/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "ssft.h"

DirList::DirList()
    : m_Md(NULL),
      mFlags(0)
#ifdef WIN32    
, mHandle(INVALID_HANDLE_VALUE)
#else
, mDirp(NULL)
#endif

{
}

DirList::~DirList()
{
    Reset();
}

void DirList::Reset()
{
    CloseDir();
    mFlags = 0;
    mIsVirgin = true;
    mCurrentPath = "";
    mIsEof = true;
}

bool DirList::IsEof() const
{
    return mIsEof;
}

void DirList::CloseDir()
{
#ifdef WIN32
    if (INVALID_HANDLE_VALUE != mHandle)
    {
        ::FindClose(mHandle);
	mHandle = INVALID_HANDLE_VALUE;
    }
#else
    if (NULL != mDirp)
    {
	closedir(mDirp);
	mDirp = NULL;
    }
#endif
}

void DirList::Open(const char *path, 
        int flags,
        const char *rootPath,
        const char *hashDigest) 
        throw(SsftDisconnect, TransferException)
{
    Reset();

    if (rootPath)
        mRootPath = rootPath;
    if (hashDigest)
        mHashDigest = hashDigest;

    std::string my_path = path;
    mCurrentPath = my_path;
    mFlags = flags;

    if (!is_dir(my_path))
      throw TransferException("Not a directory", SSFT_RESULT_NOT_A_DIRECTORY);

#ifdef WIN32
    my_path = merge_paths(my_path, std::string("\\*"));
    mHandle = FindFirstFile(my_path.c_str(), &mData);
    if (mHandle == INVALID_HANDLE_VALUE)
    {
        log << LOG_DATE
            << "DirList::Open(): "
            << "FindFirstFile(" << my_path << ") failed with error code "
            << (int)GetLastError()
            << ENDL;

        throw TransferException("Fsys error; failed to list the directory.", SSFT_RESULT_UNSPECIFIED_ERROR);
    }
#else

    mDirp = opendir(my_path.c_str());
    if (NULL == mDirp)
    {
	log << LOG_DATE
            << "DirList::Open(): "
            << "FindFirstFile(" << my_path << ") failed with error code "
            << (int)errno
            << ENDL;

        throw TransferException("Fsys error; failed to list the directory.", SSFT_RESULT_UNSPECIFIED_ERROR);
    }

#endif

    mVisitedDirs.insert(std::string(path));

    mIsEof = false;
}

std::string DirList::GetFakePath()
{
    std::string rval = mCurrentPath.c_str() + mRootPath.size();
    if (rval.empty())
        rval = "/";
#ifdef WIN32
    rval.c_str();
    char *p = &rval[0];
    while(*p)
    {
        if (*p == '\\')
            *p = '/';

        ++p;
    }
#endif

    return rval;
}

void DirList::GetPart(XmitBuffer& buffer, size_t preffered_size) 
throw(SsftDisconnect, TransferException)
{
again:
    if (!mIsEof)
    {   
        if (mIsVirgin)
        {
            std::string fake_path = GetFakePath();
            buffer << TAG_DIR_HDR;
            buffer << fake_path;
            mIsVirgin = false;
        }
        
        while(buffer.GetSize() < preffered_size)
        {
            if (!AddNodeToBuffer(buffer))
            {
                if ((mFlags & DIRFLAG_RECURSE)
                    && !mDirsToVisit.empty())
                {
                    std::string new_path = *mDirsToVisit.begin();
                    mDirsToVisit.erase(mDirsToVisit.begin());
                    
                    try
                    {
                        Open(new_path.c_str(), mFlags);
                    }
                    catch(...)
                    {
                    }

                    goto again;
                }
                
                buffer << TAG_END;
                break;
            }
        }
    }
}

void DirList::GetHash(std::string& hash, const char *name)
{
    std::string full_path = merge_paths(mCurrentPath, std::string(name));
    hash = "";

    if (is_file(full_path))
    {
#ifdef WIN32
        HANDLE fh = ::CreateFile(full_path.c_str(),
            GENERIC_READ,
            FILE_SHARE_READ | FILE_SHARE_WRITE,
            NULL,
            OPEN_EXISTING,
            FILE_FLAG_SEQUENTIAL_SCAN,
            NULL);

        if (INVALID_HANDLE_VALUE != fh)
        {
#else
        int fd = ssft_open(full_path.c_str(), O_RDONLY | O_BINARY);
        if (-1 != fd)
        {
#endif
            char buf[1024 * 16];
            EVP_MD_CTX mdctx;
            unsigned char md_value[EVP_MAX_MD_SIZE];
            unsigned int md_len = 0, i;

            memset(md_value, 0, sizeof(md_value));
            memset(&mdctx, 0, sizeof(mdctx));
            EVP_DigestInit(&mdctx, m_Md);

#ifdef WIN32
            DWORD bytes = 0;
            while(true)
            {
                if (!::ReadFile(fh, buf, sizeof(buf), &bytes, NULL))
                {
                    int err = ::GetLastError();

                    if (ERROR_HANDLE_EOF == err)
                    {
                        if (0 == bytes)
                            break; // Done
                    }
                    else
                    {
                        ::CloseHandle(fh);
                        return; // Error
                    }
                }

                if (0 == bytes)
                    break; // Done
#else
	    int bytes;
            while((bytes = read(fd, buf, sizeof(buf))) != 0)
            {
		if (bytes < 0)
		{
		    close(fd);
		    return; // error
		}
#endif

                EVP_DigestUpdate(&mdctx, buf, bytes);
            }

#ifdef WIN32
            ::CloseHandle(fh);
#else
            close(fd);
#endif

            EVP_DigestFinal(&mdctx, md_value, &md_len);

            char *p = buf;
            for(i = 0; i < md_len; i++)
            {
                p += sprintf(p, "%02x", md_value[i]);
            }

            *p = 0;

            hash = buf;
        }
    }
}

bool DirList::AddNodeToBuffer(XmitBuffer& buffer) 
throw(SsftDisconnect, TransferException)
{
    if (mIsEof)
        throw TransferException("Program error: DirList::AddNodeToBuffer(): The dir-listing has already reached EOF!", SSFT_RESULT_UNSPECIFIED_ERROR);

    bool rval = true;
    DirNodeTypesE node_type = TYPE_FILE;
    ssft_date_t mtime = 0;
    ssft_flen_t len = 0;

#ifdef WIN32
#   define NAME mData.cFileName
    len = mData.nFileSizeLow | (mData.nFileSizeHigh << 32);
    
    if (mData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
        node_type = TYPE_DIR;

    FileTimeToSsftTime(mData.ftLastWriteTime, mtime);
    { // indent
#else
#   define NAME pent->d_name

 again:
    struct dirent *pent = readdir(mDirp);

    if (NULL != pent) 
    {
	std::string new_path = merge_paths(mCurrentPath, std::string(NAME));
	ssft_stat_t st;

	if (ssft_stat(new_path.c_str(), &st))
	    goto again;

	if (S_ISDIR(st.st_mode))
	    node_type = TYPE_DIR;
	else if (S_ISREG(st.st_mode))
	    node_type = TYPE_FILE;
	else
	    goto again; // We don't want to mess with links and devices!
       
	mtime = st.st_mtime * 1000;
	len = st.st_size;
	
#endif
	std::string hash_val;
	if (NULL != m_Md)
	    GetHash(hash_val, NAME);

	buffer << TAG_NODE;
	buffer << NAME;
	buffer << node_type;
	buffer << len;
	buffer << mtime;
	buffer << hash_val;

	if ((TYPE_DIR == node_type)
	     && (mFlags & DIRFLAG_RECURSE))
	{
	    if (strcmp(NAME, ".") && strcmp(NAME, ".."))
	    {
            
#ifdef WIN32
		std::string name = mData.cFileName;
		std::string new_path = merge_paths(mCurrentPath, name, '/');
#endif
		if (mVisitedDirs.find(new_path) == mVisitedDirs.end())
		    mDirsToVisit.insert(new_path);
	    }
	}
    }
#ifdef WIN32
    rval = ::FindNextFile(mHandle, &mData) != 0;
#else
    rval = (NULL != pent);
#endif
    if (!rval)
    {
	    CloseDir();
        mIsEof = true;
    }
    
    return rval;
}

void DirList::GetList(fileset_t& dstList)
throw(SsftDisconnect, TransferException)
{
    if (mIsEof)
        return;
again:
#ifdef WIN32    
    do
    {
#else
    struct dirent *pent;
    while((pent = readdir(mDirp)) != NULL)
    {
#endif
        if (!strcmp(NAME, ".") || !strcmp(NAME, ".."))
            continue;

        FileInfo node;

#ifdef WIN32
        node.mType = ((mData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
            ? TYPE_DIR : TYPE_FILE;
        node.mLength = mData.nFileSizeLow | (mData.nFileSizeHigh << 32);
        FileTimeToSsftTime(mData.ftLastWriteTime, node.mDate);
#else
        ssft_stat_t st;
        std::string new_path = merge_paths(mCurrentPath, std::string(NAME));
        if (ssft_stat(new_path.c_str(), &st) != 0)
	        goto again;

        if (S_ISDIR(st.st_mode))
	        node.mType = TYPE_DIR;
        else if (S_ISREG(st.st_mode))
	        node.mType = TYPE_FILE;
        else
	        goto again; // We don't want to mess with links and devices!
   
        node.mDate = st.st_mtime * 1000;
        node.mLength = st.st_size;

#endif
        node.mName = merge_paths(GetFakePath(), NAME, '/');
        
        if ((NULL != m_Md) && (TYPE_FILE == node.mType))
            GetHash(node.mHash, NAME);

        dstList.insert(node);
        
        if ((TYPE_DIR == node.mType) && (mFlags & DIRFLAG_RECURSE))
        {
#ifdef WIN32
            std::string name = mData.cFileName;
            std::string new_path = merge_paths(mCurrentPath, name, '/');
#endif
            if (mVisitedDirs.find(new_path) == mVisitedDirs.end())
                mDirsToVisit.insert(new_path);
        }
#ifdef WIN32
    } while(::FindNextFile(mHandle, &mData));
#else
    }
#endif
    
    CloseDir();

next_entry:
    if ((mFlags & DIRFLAG_RECURSE)
        && !mDirsToVisit.empty())
    {
        std::string new_path = *mDirsToVisit.begin();
        mDirsToVisit.erase(mDirsToVisit.begin());
        
        try
        {
            Open(new_path.c_str(), mFlags);
        }
        catch(...)
        {
            goto next_entry;
        }
        
        goto again;
    }

    mIsEof = true;
}

