/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "ssft.h"
#include "assert.h"

using namespace std;

SvrConnection::SvrConnection(SOCKET conn, 
                             SSL_CTX *ctx,
                             const sockaddr_in *pCliAddr)
: client_addr(*pCliAddr),
mSocket(conn),
mState(STATE_CREATED),
mpCtx(ctx),
mpSsl(NULL),
mpClientCert(NULL),
mPendingBytesIn(0),
mCurrentBytesReceived(0),
mFile(-1),
mCurrentBytesXmitted(0)
{
}

SvrConnection::~SvrConnection()
{
    if (NULL != mpClientCert)
    {
        X509_free(mpClientCert);
        mpClientCert = NULL;
    }

    if (NULL != mpSsl)
    {
        SSL_free(mpSsl);
        mpSsl = NULL;
    }

    if (-1 != mSocket)
    {
        closesocket(mSocket);
        mSocket = -1;
    }

    CloseFile(-1);
}

// Returns 0 = OK, 1 = wait, -1 = error
int SvrConnection::SslAccept()
{   
    int result = SSL_accept(mpSsl);
    SslError ssl_err(mpSsl, result);

    if (debug_flag)
    {
        log << LOG_DATE 
            << "SslAccept(): "
            << "Called. Result="
            << result
            << " SSL status: "
            << ssl_err
            << ENDL;
    }

    if (0 == result)
    {
        // Shutdown
        log << LOG_DATE 
            << "ERROR - SSL_accept() failed " 
            << "The connection was shut down during handshake "
            << ssl_err 
            << ENDL;
        
        return -1;
    }
    else if (0 > result)
    {   
        if ((SSL_ERROR_WANT_READ == ssl_err.GetErrorCode())
            || (SSL_ERROR_WANT_WRITE == ssl_err.GetErrorCode()))
            return 1;
        
        log << LOG_DATE 
            << "ERROR - SSL_accept() failed " 
            << ssl_err 
            << ENDL;
        return -1;
    }
    
    SetState(STATE_CONNECTED);

    if ((0 == nocert_flag) && (NULL == mpClientCert))
    {
        mpClientCert = SSL_get_peer_certificate(mpSsl);
        if (NULL == mpClientCert)
        {
            log << LOG_DATE 
                << "The client did not provide a certificate. "
                "Disconnecting."
                << ENDL;
            return -1;
        }
    }

    // Get user rights and properties
    bool have_user = false;
    for(userptr_list_t::iterator P = user_list.begin()
        ; P != user_list.end()
        ; ++P)
    {
        if (*(*P) == mpClientCert)
        {
            // Assign the user
            mUser = *(*P);
            have_user = true;

            if (debug_flag)
            {
                log << LOG_DATE
                    << "Found matching user profile: \""
                    << mUser.GetAliasName()
                    << "\". root-path=\""
                    << mUser.GetRootPath()
                    << "\". perms=\""
                    << mUser.GetPerms()
                    << "\". "
		    << mUser.DebugInfo()
                    << ENDL;
            }
            break;
        }
    }

    if (!have_user)
    {
        char buf[256];
        X509_NAME_oneline(X509_get_subject_name(mpClientCert), buf, sizeof(buf));

        log << LOG_DATE
            << "Could not find a matching user profile for \""
            << buf
            << "\". Login denied."
            << ENDL;
        
        return -1;
    }

    // Send header
    if (debug_flag)
        log << LOG_DATE << "Sending server tag + version" << ENDL;

    char hdr[24];
    strcpy(hdr,SSFT_SVR_HELO);
    *(unsigned int *)(hdr + strlen(SSFT_SVR_HELO)) = htonl(SSFT_VERSION_1);
    try
    {
        Write(hdr, sizeof(hdr));
    }
    catch(SsftWouldBlock)
    {
    }
    catch(...)
    {
        return -1;
    }
    return 0;
}

bool SvrConnection::Create()
{
    mpSsl = SSL_new(mpCtx);
    
    if (NULL == mpSsl)
    {
        SslError ssl_err;
     
        log << LOG_DATE 
            << "ERROR - SSL_new() failed " 
            << ssl_err << ENDL;
        return false;
    }
    
    SSL_set_fd(mpSsl, mSocket);
    
    if (SslAccept() == -1)
        return false;

    return true;
}

const char *SvrConnection::GetStateName() const
{
    switch(mState)
    {
    case STATE_CREATED:
        return "STATE_CREATED";
    case STATE_CONNECTED:
        return "STATE_CONNECTED";
    case STATE_WANT_COMMAND:
        return "STATE_WANT_COMMAND";
    case STATE_GOT_CMD_HEADER_LEN:
        return "STATE_GOT_CMD_HEADER_LEN";
    case STATE_GOT_CMD_HEADER:
        return "STATE_GOT_CMD_HEADER";
    case STATE_RECEIVING:
        return "STATE_RECEIVING";
    case STATE_SENDING:
        return "STATE_SENDING";
    default:
        return "**Invalid state value **";
    }
}

const char *SvrConnection::GetCmdName() const
{
    switch(mCurrentCommand)
    {
    case SSFT_CMD_QUIT:
        return "Quit";
        
    case SSFT_CMD_TRANSF_CLI_TO_SEVER:
        return "Transfer from Client to Server";

    case SSFT_CMD_TRANSF_SEVER_TO_CLI:
        return "Transfer from Server to Client";

    case SSFT_CMD_DIR_SEVER_TO_CLI:
        return "Send a directory-listing from the Server to the Client";

    case SSFT_CMD_DELETE_FILE_OR_DIR:
        return "Delete a file or directory on the server";

    case SSFT_CMD_CREATE_DIR:
        return "Create a directory on the server";

    default:
        return "SvrConnection::GetCmdName(): Unknown command ID";
    }
}

std::string SvrConnection::Explain() const
{
    char explanation[128];

    sprintf(explanation, "%s:%u [%s]",
        inet_ntoa(client_addr.sin_addr),
        ntohs(client_addr.sin_port),
        GetStateName());

    return explanation;
}

void SvrConnection::SetState(StateE newState)
{
    mState = newState;

    if (debug_flag)
        log << LOG_DATE << "Changing state to " << GetStateName() << ENDL;
}

void SvrConnection::WriteResult(const SsftResultCodesE result)
{
    XmitBuffer hdr;
    hdr << (int)result;

    if (debug_flag)
    {
        log << LOG_DATE
            << "Replying with code "
            << result
            << ". ("
            << hdr.GetSize()
            << " bytes in reply header),"
            << ENDL;
    }

    Write(hdr.GetSendBuffer(), hdr.GetSize());
}

// Return true if data was read, else false
bool SvrConnection::Read() throw(SsftDisconnect)
{
    bool did_read = false;

    while (mCurrentBytesReceived < mPendingBytesIn)
    {
        int bytes = SSL_read(mpSsl, 
            mInputBuffer.GetRecvBuffer(mPendingBytesIn) + mCurrentBytesReceived,
            mPendingBytesIn - mCurrentBytesReceived);

        if (debug_flag)
        {
            log << LOG_DATE 
                << "Read("
                << client_addr
                << "): Read "
                << bytes
                << " bytes."
                << ENDL;
        }

        int status = SSL_get_error(mpSsl, bytes);
        switch(status)
        {
        case SSL_ERROR_NONE:
            did_read = true;
            mCurrentBytesReceived += bytes;
            continue;
        case SSL_ERROR_WANT_READ:
            goto done;
        case SSL_ERROR_ZERO_RETURN:
            if (did_read)
                goto done;
            throw SsftDisconnect(SSFT_RESULT_OPERATION_OK, "End of input stream");
            break;
        default:
            {
                SslError ssl_err(mpSsl, bytes);
                
                int err = GET_SOCKET_ERROR;
                if ((EWOULDBLOCK == err)
                    || (SSL_ERROR_WANT_READ == ssl_err.GetErrorCode())
                    || (SSL_ERROR_WANT_WRITE == ssl_err.GetErrorCode()))
                    return false; // No more to read
                
                log << LOG_DATE
                    << "Socket error on read " 
                    << err
                    << " (ssl error: "
                    << ssl_err
                    << ")"
                    << ENDL;
                
                if ((ssl_err.GetErrorCode() == 0)
                    && (0 == err))
                    return false; // Shit happens...
                
                throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR, 
                    "Socket error on read");
            }
            break;
        }
    }

done:
    return did_read;
}

void SvrConnection::OnRead() throw(SsftDisconnect)
{
    do
    {
        switch(GetState())
        {
        case STATE_CREATED:
            // Not yet initialized
            switch(SslAccept())
            {
            case 0: // OK, fall trough
                mCurrentBytesReceived = 0;
                mPendingBytesIn = 0;
                break;
            case 1: // Wait
                return;
            default: // failed
                throw SsftDisconnect(SSFT_RESULT_HANDSHAKE_FAILED, 
                    "SslAccept() failed");
            }

            // fall trough

        case STATE_CONNECTED:
            // Get the first header
            if (0 == mCurrentBytesReceived)
            {
                mPendingBytesIn = 8; // Need 8 bytes. 
            }
            Read();
            
            if (8 == mCurrentBytesReceived)
            {
                // Got the first header.
                if (strncmp(mInputBuffer.GetRecvBuffer(4), "SSFT", 4) != 0)
                {
                    // Bad header
                    WriteResult(SSFT_RESULT_BAD_PROTOCOL);
                    throw SsftDisconnect(SSFT_RESULT_BAD_PROTOCOL, 
                        "invalid initial header tag from client");
                }
                
                mClientVersion = (SsftProtocolVersionsE)ntohl(
                    *(unsigned int *)(mInputBuffer.GetRecvBuffer(8) + 4));

                switch(mClientVersion)
                {
                case SSFT_VERSION_1:
                    // We can go directly to command-loop
                    SetState(STATE_WANT_COMMAND);
                    WriteResult(SSFT_RESULT_OPERATION_OK);
                    mPendingBytesIn = mCurrentBytesReceived = 0;
                    break;
                default:
                    WriteResult(SSFT_RESULT_BAD_PROTOCOL);
                    throw SsftDisconnect(SSFT_RESULT_BAD_PROTOCOL, 
                        "invalid initial header version from client");
                }
            }
            break;

        case STATE_WANT_COMMAND:
            // Get the command opcode and header length field.
            if (0 == mPendingBytesIn)
            {
                mPendingBytesIn = 4; // header length field
                mInputBuffer.Reset();
                Read();
            }

            if (4 == mCurrentBytesReceived)
            {
                mInputBuffer >> mPendingBytesIn; // Header length, including the header length field

                if (mPendingBytesIn < 4)
                {
                    throw SsftDisconnect(SSFT_RESULT_INVALID_HEADER, 
                        "Error: Invalid (empty) header from client.");
                }

                SetState(STATE_GOT_CMD_HEADER_LEN);
            }

            if (mPendingBytesIn == mCurrentBytesReceived)
                throw SsftDisconnect(SSFT_RESULT_INVALID_HEADER, 
                        "Program error: Detected dead loop. Terminating the connection.");

            break;

            // Read the complete header
        case STATE_GOT_CMD_HEADER_LEN:

            Read();
            
            if (mPendingBytesIn == mCurrentBytesReceived)
            {  
                // Got header
                mCurrentBytesReceived = 0;
                int tmp_cmd = 0;
                mInputBuffer >> tmp_cmd;
                mCurrentCommand = (SsftCmdE)tmp_cmd;
                SetState(STATE_GOT_CMD_HEADER);

                if (verbose_flag)
                {
                    log << LOG_DATE 
                        << "Client sent "
                        <<  GetCmdName()
                        << " command"
                        << ENDL;
                }
            }
            else
                break;

            // fall trough
            
        case STATE_GOT_CMD_HEADER:
            switch(mCurrentCommand)
            {
            case SSFT_CMD_QUIT:
                WriteResult(SSFT_RESULT_OPERATION_OK);
                throw SsftDisconnect(SSFT_RESULT_OPERATION_OK, 
                    "QUIT command from client");
                break;
            case SSFT_CMD_CREATE_DIR:
                {
                    if (!mUser.HaveRight("createdir"))
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected mkdir due to missing \"createdir\" privilege."
                            << ENDL;
                        
                        continue;
                    }

                    int my_flags = 0;
                    mCurrentBytesReceived = 0;
                    mPendingBytesIn = 0;
                    std::string hash_name;

                    mInputBuffer >> mCurrentDstPath;
                    mInputBuffer >> my_flags;
                    mInputBuffer >> hash_name;

                    if (!PrcName(false))
                        continue;

                    WriteResult(CreateDir(my_flags));
                    SetState(STATE_WANT_COMMAND);
                    continue;
                }
                break;
            case SSFT_CMD_DELETE_FILE_OR_DIR:
                {
                    if (!mUser.HaveRight("delete"))
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected delete due to missing \"delete\" privilege."
                            << ENDL;
                        
                        continue;
                    }

                    
                    int my_flags = 0;
                    mCurrentBytesReceived = 0;
                    mPendingBytesIn = 0;
                    std::string hash_name;

                    mInputBuffer >> mCurrentDstPath;
                    mInputBuffer >> my_flags;
                    mInputBuffer >> hash_name;

                    if (!PrcName(false))
                        continue;

                    WriteResult(DeletePath(my_flags));
                    SetState(STATE_WANT_COMMAND);
                    continue;
                }
                break;
            case SSFT_CMD_DIR_SEVER_TO_CLI:
                {
                    if (!mUser.HaveRight("list"))
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected dir-list due to missing \"list\" privilege."
                            << ENDL;
                        
                        continue;
                    }

                    int my_flags = 0;
                    mCurrentBytesReceived = 0;
                    mPendingBytesIn = 0;
                    std::string hash_name;

                    mInputBuffer >> mCurrentDstPath;
                    mInputBuffer >> my_flags;
                    mInputBuffer >> hash_name;

                    // Check hash
                    if (hash_name.empty())
                        mDir.m_Md = NULL;
                    else if ((mDir.m_Md = EVP_get_digestbyname(hash_name.c_str()))
                        == NULL)
                    {
                        WriteResult(SSFT_RESULT_UNKNOWN_HASH_DIGEST);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected dir-list due to unknown hash-digest name: "
                            << hash_name
                            << ENDL;
                        
                        continue;
                    }

                    if (!PrcName(false))
                        continue;

                    try
                    {
                        mDir.Open(mCurrentFilePath.c_str(), 
                            my_flags,
                            mUser.GetRootPath().c_str(),
                            hash_name.c_str());
                    }
                    catch(TransferException& e)
                    {
                        WriteResult(e.GetResult());
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected dir-list due to error: "
                            << e.what()
                            << ENDL;
                        
                        continue;
                    }


                    // Send header (ok, size and date)
                    XmitBuffer hdr;
                    hdr << SSFT_RESULT_OPERATION_OK;
                    Write(hdr.GetSendBuffer(), hdr.GetSize());

                    // Start sending the file
                    SetState(STATE_SENDING);
                    mCurrentBytesXmitted = 0;

                    log << LOG_DATE
                        << "Sending dir-lsting to client "
                        << mUser.GetAliasName()
                        << " ("
                        << client_addr
                        << "): name=\""
                        << mCurrentFilePath
                        << "\"."
                        << ENDL;

                    goto sending;
                }
                break;
                
            case SSFT_CMD_TRANSF_SEVER_TO_CLI:
                {
                    mCurrentBytesReceived = 0;
                    mPendingBytesIn = 0;

                    if (!mUser.HaveRight("read"))
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected download due to missing \"read\" privilege."
                            << ENDL;
                        
                        continue;
                    }

                    mInputBuffer >> mCurrentDstPath;
                    mCurrentFileName = "";

                    if (!PrcName(false))
                        continue;
                    
                    // Stat the file
                    ssft_stat_t st;
                    int result = ssft_stat(mCurrentFilePath.c_str(), &st);
                    if (0 != result)
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected download. The target 'file' \""
                            << mCurrentFilePath
                            << "\" can not be accessed (stat() failed)."
                            << ENDL;
                        
                        continue;
                    }

                    if ((st.st_mode & S_IFREG) != S_IFREG)
                    {
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected download. The target 'file' \""
                            << mCurrentFilePath
                            << "\" is not a plain file."
                            << ENDL;
                        
                        continue;
                    }
                    
                    // See if we can open the file
                    if (-1 != mFile)
                        close(mFile); // Should never happen
                    mFile = ssft_open(mCurrentFilePath.c_str(), O_RDONLY | O_BINARY);
                    if (0 > mFile)
                    {
                        mCurrentBytesReceived = 0;
                        mPendingBytesIn = 0;
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected download. The target 'file' \""
                            << mCurrentFilePath
                            << "\" could not be opened."
                            << ENDL;
                        
                        continue;
                    }
                    
                    // Send header (ok, size and date)
                    XmitBuffer hdr;
                    hdr << SSFT_RESULT_OPERATION_OK;
                    hdr << (ssft_flen_t)st.st_size;
                    hdr << (ssft_date_t)st.st_mtime * 1000;
                    Write(hdr.GetSendBuffer(), hdr.GetSize());

                    // Start sending the file
                    SetState(STATE_SENDING);
                    mCurrentBytesXmitted = 0;
                    mCurrentFileLength = st.st_size;
                    mCurrentFileDate = st.st_ctime;

                    log << LOG_DATE
                        << "Sending file to client "
                        << mUser.GetAliasName()
                        << " ("
                        << client_addr
                        << "): name=\""
                        << mCurrentFilePath
                        << "\" size="
                        << mCurrentFileLength
                        << " bytes."
                        << ENDL;
                    
                    if (mCurrentFileLength == 0)
                    {
                        CloseFile(0);
                        
                        log << LOG_DATE
                            << "Successfully sent empty file \""
                            << mCurrentFilePath
                            << "\" ("
                            << mCurrentFileLength
                            << " bytes) to client "
                            << mUser.GetAliasName()
                            << " on "
                            << client_addr
                            << ENDL;
                        
                        SetState(STATE_WANT_COMMAND);
                        mCurrentBytesXmitted = 0;
                        mCurrentBytesReceived = 0;
                        mPendingBytesIn = 0;
                    }
                    else
                        goto sending;
                }
                break;

            case SSFT_CMD_TRANSF_CLI_TO_SEVER:
                {
                    if (!mUser.HaveRight("write"))
                    {
                        mCurrentBytesReceived = 0;
                        mPendingBytesIn = 0;
                        WriteResult(SSFT_RESULT_ACCESS_DENIED);
                        SetState(STATE_WANT_COMMAND);

                        log << LOG_DATE
                            << "User "
                            << mUser.GetAliasName()
                            << ": Rejected upload due to missing \"write\" privilege."
                            << ENDL;
                        
                        continue;
                    }
                    // Parse down header
                    mInputBuffer >> mCurrentFileName;
                    mInputBuffer >> mCurrentDstPath;
                    mInputBuffer >> mCurrentFileDate;
                    mInputBuffer >> mCurrentFileLength;

                    if (!PrcName())
                        continue;
                    
                    if (debug_flag)
                    {
                        log << LOG_DATE
                            << "Physical file is: \""
                            << mCurrentFilePath
                            << "\"."
                            << ENDL;
                    }

                    // Open file
                    assert(-1 == mFile);
                    mFile = ssft_open(mCurrentFilePath.c_str(), 
                        O_CREAT | O_SEQUENTIAL | O_TRUNC | O_WRONLY | O_BINARY,
                        S_IREAD | S_IWRITE);
                    
                    if (-1 == mFile)
                    {
                        log << LOG_DATE 
                            << "Failed to open file \""
                            << mCurrentFilePath
                            << "\" for write. Error code: "
                            << errno
                            << " "
                            << strerror(errno)
                            << ENDL;
                        
                        WriteResult(SSFT_RESULT_FAILED_TO_OPEN_FILE);
                        throw SsftDisconnect(SSFT_RESULT_FAILED_TO_OPEN_FILE, 
                                "Failed to open file for write");
                    }

                    set_ownership(mCurrentFilePath, 
                        mUser.GetUnixModeFile(),
                        mUser.GetOwnerUser().c_str(), 
                        mUser.GetOwnerGroup().c_str());
                    
                    // Start the transfer
                    WriteResult(SSFT_RESULT_OPERATION_OK);
                    mPendingBytesIn = 0;
                    mCurrentBytesReceived = 0;
                    SetState(STATE_RECEIVING);
                    
                    log << LOG_DATE
                        << "Receiving file from client "
                        << mUser.GetAliasName()
                        << " ("
                        << client_addr
                        << "): name=\""
                        << mCurrentFilePath
                        << "\" size="
                        << mCurrentFileLength
                        << " bytes."
                        << ENDL;
                    
                    
                   if (0 == mCurrentFileLength)
                   {
                       // Empty file
                       OnTransferOk();
                   }
                }
                break;
            }
            
            break;
            
        case STATE_SENDING:
sending:
            {
                OnWrite();
            }
            break;
            
        case STATE_RECEIVING:
            {
                int bytes = 0;
                bool loop_on_read = true;
                while(loop_on_read)
                {
                    char buf[SSFT_NET_BUFLEN];
                    bytes = SSL_read(mpSsl, buf, sizeof(buf));
                    int status;
                    
                    switch(status = SSL_get_error(mpSsl, bytes))
                    {
                    case SSL_ERROR_NONE:
                        mCurrentBytesReceived += bytes;
                        mCurrentBytesXmitted += bytes;
                        
                        if (mCurrentBytesReceived > mCurrentFileLength)
                        {
                            log << LOG_DATE 
                                << "File too big \""
                                << mCurrentFileName.c_str()
                                << "File len agreed: "
                                << mCurrentBytesReceived
                                << ". Actual file len: "
                                << mCurrentBytesReceived
                                << " (or more)."
                                << ENDL;
                            
                            CloseFile(-1);
                            WriteResult(SSFT_RESULT_FILE_TOO_LONG);
                            
                            throw SsftDisconnect(SSFT_RESULT_FILE_TOO_LONG, 
                                "File too long");
                        }
                        
                        if (write(mFile, buf, bytes) != bytes)
                        {
                            log << LOG_DATE 
                                << "Write error on file \""
                                << mCurrentFileName.c_str()
                                << "\" for write. Error code: "
                                << errno
                                << " "
                                << strerror(errno);
                            
                            CloseFile(errno);
                            WriteResult(SSFT_RESULT_WRITE_ERROR);
                            
                            throw SsftDisconnect(SSFT_RESULT_WRITE_ERROR, 
                                "Write error");
                        }
                        
                        if (mCurrentBytesReceived == mCurrentFileLength)
                        {
                            OnTransferOk();
                            loop_on_read = false;
                        }
                        break;
                        
                    case SSL_ERROR_WANT_READ:
                        loop_on_read = false;
                        break;
                        
                    case SSL_ERROR_ZERO_RETURN:
                    default:
                        {
                            // Error
                            SslError ssl_err(mpSsl, bytes);
                            
                            int err = GET_SOCKET_ERROR;                        
                            CloseFile(err);
                            throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR, 
                                "Socket error on read");
                        }
                        break;
                    } 
                } 
            }
        }
    } while(Read());
}

void SvrConnection::OnTransferOk()
{
    CloseFile(0);
    
    mCurrentBytesReceived = 0;
    mPendingBytesIn = 0;
    WriteResult(SSFT_RESULT_OPERATION_OK);
    SetState(STATE_WANT_COMMAND);
    
    log << LOG_DATE
        << "Received file client ("
        << client_addr
        << "): name=\""
        << mCurrentFileName.c_str()
        << "\" size="
        << mCurrentFileLength
        << " bytes. OK"
        << ENDL;
}

bool SvrConnection::Write(const char *buffer, size_t length)
{
    size_t pos = mOutputBuffer.size();
    mOutputBuffer.resize(pos + length);
    memcpy(&mOutputBuffer[pos], buffer, length);

    if (debug_flag)
    {
        log << LOG_DATE 
            << "Write: Writing "
            << length
            << " bytes."
            << ENDL;
    }

    assert(mOutputBuffer.size() >= length);

    try
    {
        OnWrite();
    }
    catch(SsftWouldBlock)
    {
        return false;
    }

    return true;
}

void SvrConnection::OnWrite() throw(SsftWouldBlock, SsftDisconnect)
{
again:
    while(!mOutputBuffer.empty())
    {
        int bytes = SSL_write(mpSsl, &mOutputBuffer[0], mOutputBuffer.size());
        int status = SSL_get_error(mpSsl, bytes);
        switch(status)
        {
        case SSL_ERROR_NONE:
            if (bytes == (int)mOutputBuffer.size())
                mOutputBuffer.resize(0);
            else
            {
                memmove(&mOutputBuffer[0], &mOutputBuffer[bytes], 
                    mOutputBuffer.size() - bytes);
                mOutputBuffer.resize(mOutputBuffer.size() - bytes);
            }
            break;
        case SSL_ERROR_WANT_WRITE:
            throw SsftWouldBlock();
            break;
        default:
            {
                int socket_err = GET_SOCKET_ERROR;
                if (EWOULDBLOCK == socket_err)
                    throw SsftWouldBlock();
                
                SslError ssl_err;
                
                log << LOG_DATE 
                    << "ERROR - OnWrite() failed " 
                    << ssl_err 
                    << " Socket error: " << socket_err
                    << ENDL;
                
                throw SsftDisconnect(SSFT_RESULT_NETWORK_ERROR);
            }
            break;
        }
    }

    if (GetState() == STATE_SENDING)
    {
        if (SSFT_CMD_DIR_SEVER_TO_CLI == mCurrentCommand)
        {
            XmitBuffer buf;

            if (mDir.IsEof())
                goto end_of_file;
            
            mDir.GetPart(buf);
            mOutputBuffer = buf.GetSendVector();
        }
        else if (mCurrentFileLength == mCurrentBytesXmitted)
        {
            // EOT
            CloseFile(0);

            log << LOG_DATE
                << "Successfully sent file \""
                << mCurrentFilePath
                << "\" ("
                << mCurrentFileLength
                << " bytes) to client "
                << mUser.GetAliasName()
                << " on "
                << client_addr
                << ENDL;

end_of_file:
            SetState(STATE_WANT_COMMAND);
            mCurrentBytesXmitted = 0;
            mCurrentBytesReceived = 0;
            mPendingBytesIn = 0;
        }
        else
        {
            mOutputBuffer.resize(1024);
            int bytes = read(mFile, &mOutputBuffer[0], mOutputBuffer.size());
            if (0 >= bytes)
            {
                log << LOG_DATE
                    << "Read error on file \""
                    << mCurrentFilePath
                    << "\" : "
                    << strerror(errno)
                    << ". Transfer to client "
                    << mUser.GetAliasName()
                    << " on "
                    << client_addr
                    << " aborted after "
                    << mCurrentBytesXmitted
                    << " of "
                    << mCurrentFileLength
                    << " bytes."
                    << ENDL;

                close(mFile);
                mFile = -1;
                throw SsftDisconnect(SSFT_RESULT_READ_ERROR);
            }

            mCurrentBytesXmitted += bytes;
            if (mOutputBuffer.size() != (size_t)bytes)
                mOutputBuffer.resize(bytes);
            goto again;
        }
    }
}

void SvrConnection::CloseFile(int error)
{
    if (-1 != mFile)
    {
        if (debug_flag)
        {
            log << LOG_DATE 
                << "Closing file : \""
                << mCurrentFileName.c_str()
                << "\" (client "
                << client_addr
                << "). Error code is "
                << error
                << " "
                << strerror(error)
                << ". Bytes xmitted: "
                << mCurrentBytesXmitted
                << ENDL;
        }

        close(mFile);
        mFile = -1;
        mCurrentBytesXmitted = 0;
    }
}


bool SvrConnection::PrcName(bool doCreateDirs) throw(SsftDisconnect)
{
    // Get real destination name
    
    if (mUser.GetRootPath().empty())
    {
        WriteResult(SSFT_RESULT_ACCESS_DENIED);
        
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": Root-path is not configured!"
            << ENDL;
        
        throw SsftDisconnect(SSFT_RESULT_ACCESS_DENIED, 
            "Local configuration error");
    }

    mCurrentFilePath = merge_paths(mUser.GetRootPath(), 
        mCurrentDstPath);
    
    try
    {
        mCurrentFilePath = parse_path(mCurrentFilePath);
    }
    catch(exception)
    {
        // Invalid path, potential crack attempt
        
        mCurrentBytesReceived = 0;
        mPendingBytesIn = 0;
        WriteResult(SSFT_RESULT_ACCESS_DENIED);
        SetState(STATE_WANT_COMMAND);
        
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": Invalid path: \""
            << mCurrentFilePath
            << "\"."
            << ENDL;
        
        return false;
    }
    
    //if (doCreateDirs && mUser.HaveRight("createdir"))
    //    make_dirs(mCurrentFilePath)
    
    if (!mCurrentFileName.empty() && is_dir(mCurrentFilePath))
    {
        const char *p = mCurrentFileName.c_str();
        if (p)
        {
            const char *my_name = p;
            while(*p)
            {
                if (IS_SLASH(*p))
                    my_name = p + 1;
                ++p;
            }
            
            if (*my_name)
            {
                mCurrentFilePath = merge_paths(mCurrentFilePath,
                    my_name);
            }
        }
    }

    if (SSFT_MAX_PATHLEN <= mCurrentFilePath.size())
    {
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": Path too long (max "
            << SSFT_MAX_PATHLEN
            << " letters): \""
            << mCurrentFilePath
            << "\" Access denied.."
            << ENDL;
    }

#ifdef WIN32
    int i;
    while((i = mCurrentFilePath.find('/')) != mCurrentFilePath.npos)
        mCurrentFilePath[i] = '\\';
#endif

    // Last check
    try
    {
        mCurrentFilePath = parse_path(mCurrentFilePath);
    }
    catch(exception)
    {
        // Invalid path, potential crack attempt
        
        mCurrentBytesReceived = 0;
        mPendingBytesIn = 0;
        WriteResult(SSFT_RESULT_ACCESS_DENIED);
        SetState(STATE_WANT_COMMAND);
        
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": Invalid path: \""
            << mCurrentFilePath
            << "\". Access denied."
            << ENDL;
        
        return false;
    }

    const string my_root = mUser.GetRootPath();
    if ((mCurrentFilePath.size() < my_root.size())
        || strncmp(mCurrentFilePath.c_str(), my_root.c_str(),
        my_root.size())
        || !IS_SLASH(mCurrentFilePath[my_root.size()]))
    {
        mCurrentBytesReceived = 0;
        mPendingBytesIn = 0;
        WriteResult(SSFT_RESULT_ACCESS_DENIED);
        SetState(STATE_WANT_COMMAND);
        
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": The specified path goes above the users root: \""
            << mCurrentFilePath
            << "\". Access denied."
            << ENDL;
        
        return false;
    }

    if (is_device(mCurrentFilePath.c_str()))
    {
        mCurrentBytesReceived = 0;
        mPendingBytesIn = 0;
        WriteResult(SSFT_RESULT_ACCESS_DENIED);
        SetState(STATE_WANT_COMMAND);
        
        log << LOG_DATE
            << "User "
            << mUser.GetAliasName()
            << ": The specified path may be a device: \""
            << mCurrentFilePath
            << "\". Access denied."
            << ENDL;
        
        return false;
    }

     // Make dirs if required
    if (doCreateDirs && !is_dir(get_pathname(mCurrentFilePath)))
    {
        if (!mUser.HaveRight("createdir"))
        {
             log << LOG_DATE
                << "User "
                << mUser.GetAliasName()
                << ": The directory : \""
                << get_pathname(mCurrentFilePath)
                << "\". does not exist, and the user does not have the right to create directories."
                << ENDL;
            
            mCurrentBytesReceived = 0;
            mPendingBytesIn = 0;
            WriteResult(SSFT_RESULT_ACCESS_DENIED);
            SetState(STATE_WANT_COMMAND);
            return false;
        }

        make_dirs(mCurrentFilePath,
            mUser.GetUnixModeDir(),
            mUser.GetOwnerUser().c_str(),
            mUser.GetOwnerGroup().c_str());
    }

    return true;
}

SsftResultCodesE SvrConnection::CreateDir(int flags)
{
    log << LOG_DATE
        << "User "
        << mUser.GetAliasName()
        << ": Creating directory \""
        << mCurrentFilePath
        << "\""
        << ENDL;

    if (!make_dirs(with_slash(mCurrentFilePath), 
        mUser.GetUnixModeDir(),
        mUser.GetOwnerUser().c_str(),
        mUser.GetOwnerGroup().c_str()))
        return SSFT_RESULT_UNSPECIFIED_ERROR;

    return SSFT_RESULT_OPERATION_OK;
}

SsftResultCodesE SvrConnection::DeletePath(int flags, const char *path)
{
    if (NULL == path)
        path = mCurrentFilePath.c_str();

    log << LOG_DATE
        << "User "
        << mUser.GetAliasName()
        << ": Deleting directory \""
        << path
        << "\""
        << ENDL;

#ifdef WIN32
    if (is_file(path))
    {
        if (!::DeleteFile(path))
        {
            int err = ::GetLastError();
            log << LOG_DATE
                << "Failed to delete the file \""
                << path
                << "\". System error: "
                << err
                << ENDL;
            
            switch(err)
            {
            case ERROR_FILE_NOT_FOUND:
            case ERROR_PATH_NOT_FOUND:
                return SSFT_RESULT_PATH_NOT_FOUND;
            case ERROR_ACCESS_DENIED:
                return SSFT_RESULT_ACCESS_DENIED;
            case ERROR_SHARING_VIOLATION:
                return SSFT_RESULT_SHARING_ERROR;
            default:
                return SSFT_RESULT_UNSPECIFIED_ERROR;
            }
        }

        return SSFT_RESULT_OPERATION_OK;
    }
#endif
    
    if (is_dir(path))
    {
        if (flags & DIRFLAG_RECURSE)
        {
#ifdef WIN32
#           define NAME data.cFileName 
            WIN32_FIND_DATA  data;            
            std::string my_path = merge_paths(std::string(path), 
                std::string("*"));

            HANDLE h = ::FindFirstFile(my_path.c_str(), &data);
            
            if (h != INVALID_HANDLE_VALUE)
            {
                do
                {
#else
#define     NAME pent->d_name
	        DIR *pdir = ::opendir(path);
	        struct dirent *pent;
	        while((pent = ::readdir(pdir)) != NULL)
    	    {
	            ssft_stat_t st;
                std::string full_name = merge_paths(std::string(path), 
                    std::string(NAME));
		    if (ssft_stat(full_name.c_str(), &st) != 0)
		        continue;
		    if (S_ISREG(st.st_mode) || S_ISDIR(st.st_mode))
		    {
#endif
                if (strcmp(NAME, ".") 
                    && strcmp(NAME, ".."))
                {
#ifdef WIN32
                    std::string full_name = merge_paths(std::string(path), 
                        std::string(data.cFileName));
#endif               
                    SsftResultCodesE result = 
                        DeletePath(flags, full_name.c_str());

                    if (SSFT_RESULT_OPERATION_OK != result)
                    {
#ifdef WIN32
                        ::FindClose(h);
#else
			            ::closedir(pdir);
#endif
                        return result;
                    }
                }
                    
#ifdef WIN32
                } while(::FindNextFile(h, &data));

                ::FindClose(h);
            } 
#else
	            } // if (S_REG...
	        }  // while(pent ...
	    
	        ::closedir(pdir);
#endif
        }

#ifdef WIN32
        if (!::RemoveDirectory(path))
        {
            int err = ::GetLastError();
            log << LOG_DATE
                << "Failed to delete the directory \""
                << path
                << "\". System error: "
                << err
                << ENDL;

            switch(err)
            {
            case ERROR_FILE_NOT_FOUND:
            case ERROR_PATH_NOT_FOUND:
                return SSFT_RESULT_PATH_NOT_FOUND;
            case ERROR_ACCESS_DENIED:
                return SSFT_RESULT_ACCESS_DENIED;
            case ERROR_SHARING_VIOLATION:
                return SSFT_RESULT_SHARING_ERROR;
            default:
                return SSFT_RESULT_UNSPECIFIED_ERROR;
            }
        }

        return SSFT_RESULT_OPERATION_OK;
    }

    return SSFT_RESULT_PATH_NOT_FOUND;

#else
    }
    //In UNIX, we use the same system call to delete files and dirs

    if (::unlink(path) != 0)
    {
        log << LOG_DATE
            << "Failed to delete the path \""
            << path
            << "\". System error: "
            << (int)errno
            << ENDL;
        
        return SSFT_RESULT_UNSPECIFIED_ERROR;
    }

    return SSFT_RESULT_OPERATION_OK;
#endif
}



