/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/


#include "StdAfx.h"
#include "ssft.h"
#include <assert.h>

using namespace std;

User::User()
: mUnixModeFile(0), mUnixModeDir(0)
{
    CreateTokenList();
}

User::User(const char *idString)
: mUnixModeFile(0), mUnixModeDir(0)
{
    const char *p = idString;
    string sect;

    // Create token list

    CreateTokenList();

    while(*p)
    {
        sect = "";

        while(isspace(*p) || (',' == *p) || ('/' == *p))
            ++p;

        if (']' == *p)
            break; // End of relevant input

        // Get tag
        while(*p && ('=' != *p))
            sect += *p++;

        int nid = 0;
        if (sect.empty() 
            || ((nid = OBJ_txt2nid((char *)sect.c_str())) <= 0))
        {
             log << LOG_DATE
                << "Warning: Invalid ID string or bad id token in string: \""
                << idString
                << "\". The user can not be authenticated."
                << ENDL;
            return;
        }

        // Get value

        sect = "";

        if ('=' == *p)
            ++p;

        while(*p && ('/' != *p) && (',' != *p) && (']') != *p)
            sect += *p++;

        if (!sect.empty())
            mIdPatterns.push_back(IdItem(nid, sect));
    }

    if (mIdPatterns.empty())
    {
        log << LOG_DATE 
            << "Warning: User defined by \""
            << idString
            << "\" does not have any identification options. "
            "The user can not be authenticated."
            << ENDL;
    }
}

User::~User()
{
}

bool User::operator == (const X509 *psert) const
{
    iditemlist_t::const_iterator P;

    for(P = mIdPatterns.begin()
        ; P != mIdPatterns.end()
        ; ++P)
    {
        char buf[256];

        int result = X509_NAME_get_text_by_NID(X509_get_subject_name((X509 *)psert),
            P->GetNid(), buf, 256);
        if (result > 0)
        {
            // Compare
            const string & pattern_str = P->GetValue();

            if (!xstrcmp(pattern_str.c_str(), buf))
                return false;
        }
    }

    return true;
}

void User::operator = (const User& v)
{
    mUserAlias = v.mUserAlias;
    mRootPath = v.mRootPath;
    mPerms = v.mPerms;
    mIdPatterns = v.mIdPatterns;
    mUnixModeFile = v.mUnixModeFile;
    mUnixModeDir = v.mUnixModeDir;
    mOwnerUser = v.mOwnerUser;
    mOwnerGroup = v.mOwnerGroup;

    CreateTokenList();
}


void User::CreateTokenList()
{
    mTokenList.erase(mTokenList.begin(), mTokenList.end());
    mTokenList.resize(8);

    ConfToken *ptoken = &mTokenList[0];

    ptoken->mType = TOKEN_STRING;
    ptoken->mpName = "root-path";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mRootPath;

    ++ptoken;
    ptoken->mType = TOKEN_STRING;
    ptoken->mpName = "perms";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mPerms;

    ++ptoken;
    ptoken->mType = TOKEN_STRING;
    ptoken->mpName = "alias";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mUserAlias;

    ++ptoken;
    ptoken->mType = TOKEN_OCT;
    ptoken->mpName = "unix-mode-file";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mUnixModeFile;

    ++ptoken;
    ptoken->mType = TOKEN_OCT;
    ptoken->mpName = "unix-mode-dir";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mUnixModeDir;

    ++ptoken;
    ptoken->mType = TOKEN_STRING;
    ptoken->mpName = "owner-user";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mOwnerUser;

    ++ptoken;
    ptoken->mType = TOKEN_STRING;
    ptoken->mpName = "owner-group";
    ptoken->mLength = strlen(ptoken->mpName);
    ptoken->mpPtr = &mOwnerGroup;

    ++ptoken;
    ptoken->mType = TOKEN_INVALID;
    ptoken->mLength = 0;
    ptoken->mpName = NULL;
    ptoken->mpPtr = NULL;
}

bool User::HaveRight(const char *name)
{
    vector<char> tmp_name(mPerms.size() + 1);
    memcpy(&tmp_name[0], mPerms.c_str(), mPerms.size());
    bool rval = false;

    for(char *p = strtok(&tmp_name[0], " ,")
        ; p
        ; p = strtok(NULL, " ,"))
    {
        if (stricmp(name, p) == 0)
        {
            rval = true;
            break;
        }
    }
    
    return rval;
}


std::string User::DebugInfo() const
{
  std::string rval;
  const ConfToken *ptoken;

  for(ptoken = &mTokenList[0]
      ; TOKEN_INVALID != ptoken->mType
      ; ptoken++)
  {
    rval += ptoken->mpName;
    rval += "=";
    switch(ptoken->mType)
    {
    case TOKEN_STRING:
      rval += *(const std::string *)ptoken->mpPtr;
      break;
    case TOKEN_INT:
      {
	char buf[128];
	sprintf(buf, "%d", *(const int *)ptoken->mpPtr);
	rval += buf;
      }
      break;
    case TOKEN_OCT:
      {
	char buf[128];
	sprintf(buf, "%o", *(const int *)ptoken->mpPtr);
	rval += buf;
      }
      break;
    default:
      break;
    }

    rval += "; ";
  }

  return rval;
}

