/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/


#include "StdAfx.h"
#include <assert.h>
#include "ssft.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XmitBuffer::XmitBuffer()
{
    Reset();
}

XmitBuffer::~XmitBuffer()
{

}

void XmitBuffer::AddString(const char *ptr, const size_t len) throw(SsftDisconnect)
{
    AddInt32(len);

    if (len)
    {
        Resize(len);
        memcpy(&mBuffer[mCurInPos], ptr, len);
        mCurInPos += len;
    }
}

void XmitBuffer::GetString(std::string& str) throw(SsftDisconnect)
{
    size_t len;
    GetInt32(*(int *)&len);

    if (mCurOutPos > (mBuffer.size() + len))
        throw SsftDisconnect(-1, "XmitBuffer::GetString() - Buffer underflow");

    if (len)
    {
        str.resize(len);

        memcpy(&str[0], &mBuffer[mCurOutPos], len);
        mCurOutPos += len;
    }
    else
        str.empty();
}

void XmitBuffer::AddInt32(const int v) throw(SsftDisconnect)
{
    Resize(4);

    *(int *)(&mBuffer[mCurInPos]) = htonl(v);
    mCurInPos += 4;
}

void XmitBuffer::GetInt32(int& v) throw(SsftDisconnect)
{
    if (mCurOutPos > (mBuffer.size() + 4))
        throw SsftDisconnect(-1, "XmitBuffer::GetInt32() - Buffer underflow");

    v = ntohl(*(int *)(&mBuffer[mCurOutPos]));
    mCurOutPos += 4;
}

void XmitBuffer::AddInt64(const ssft_flen_t& v) throw(SsftDisconnect)
{
    Resize(8);

    *(ssft_flen_t *)(&mBuffer[mCurInPos]) = htonll(v);
    mCurInPos += 8;
}

void XmitBuffer::GetInt64(ssft_flen_t& v) throw(SsftDisconnect)
{
    if (mCurOutPos > (mBuffer.size() + 8))
        throw SsftDisconnect(-1, "XmitBuffer::GetInt32() - Buffer underflow");

    v = ntohll(*(ssft_flen_t *)(&mBuffer[mCurOutPos]));
    mCurOutPos += 8;
}

void XmitBuffer::Resize(size_t new_chunc) throw(SsftDisconnect)
{
    size_t new_len = mBuffer.size() + new_chunc;

    if (MAX_BUF_LEN < new_len)
        throw SsftDisconnect(-1, "XmitBuffer::Resize() - Buffer overflow");

    mBuffer.resize(mBuffer.size() + new_chunc);
}

size_t XmitBuffer::GetSize() const
{
    return mBuffer.size();
}

bool XmitBuffer::IsAtEndOfOutput() const
{
    return GetSize() <= mCurOutPos;
}

char *XmitBuffer::GetRecvBuffer(size_t size) throw(SsftDisconnect)
{
    if (size > mBuffer.size())
        Resize(size - mBuffer.size());

    return &mBuffer[0];
}

const char *XmitBuffer::GetSendBuffer()
{
    assert(mBuffer.size() >= 4);
    assert(mBuffer.size() == mCurInPos);

    // Set size item
    *(int *)(&mBuffer[0]) = htonl(mBuffer.size());
    
    return &mBuffer[0];
}

const std::vector<char>& XmitBuffer::GetSendVector()
{
    GetSendBuffer();
    return mBuffer;
}

void XmitBuffer::Reset()
{
    mBuffer.erase(mBuffer.begin(), mBuffer.end());
    mBuffer.reserve(MAX_BUF_LEN + sizeof(int));
    mCurInPos = 0;
    mCurOutPos = 0;
    *this << (int)0; // Reserve space for size-item
}
