/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/


#include "StdAfx.h"
#include "ssft.h"
#include <assert.h>

using namespace std;

typedef list<string> arglist_t;

// Extract the host from the list of arguments.
// It must be forst or last, and the position also
// indicates the transfer direction.
bool extract_host(int& argc, 
                  char ***argv, 
                  std::string& destPath,
                  bool& clientToServer)
{
    int host_at = -1;

    for(int index = 0; index < argc; index++)
    {
        const char *path = argv[0][index];
#ifdef WIN32
        if (isalpha(path[0]) && (path[1] == ':')) 
            path += 2;
#endif
        const char *p = strchr(path, ':');
        if (p)
        {
            if ((-1 != host_at)
                || ((index!= 0) && (index != (argc -1))))
            {
                log << LOG_DATE
                    << "Bad pathname: \""
                    << argv[0][index]
                    << "\". Illegal character ':'!"
                    << ENDL;
                return false;
            }

            host_at = index;
        }
    }

    if (-1 == host_at)
    {
        log << LOG_DATE
            << "Error parsing arguments: You have not specified a server."
            << ENDL;
        return false;
    }

    char *path = argv[0][host_at];
    while(*path && (':' != *path))
        hostname += *path++;
    
    if (*path == ':')
        ++path;
    
    if (!*path)
    {
        log << LOG_DATE 
            << "Error parsing arguments: No path after hostname! "
            << argv[0][host_at]
            << ENDL;

        return false;
    }

    --argc;
    if (0 == host_at)
    {
        // host:/ local
        clientToServer = false;

        argv[0][0] = path;
        destPath = argv[0][argc];
    }
    else
    {
        // local host:/
        destPath = path;
        clientToServer = true;
    }

    return true;
}

int do_client_xmit(int argc, char **argv)
{
    // Do a fundamental check on the argument(s)
    switch(curr_cmd)
    {
    case SSFT_MODE_COPY:
        if (argc < 2)
        {
            log << "No files specifield" << ENDL;
            return -1;
        }
        break;
    case SSFT_MODE_REPLICATE:
        if (argc != 2)
        {
            log << "Bad number of arguments. Must have two arguments "
                "for this command." << ENDL;
            return -1;   
        }
        break;
    case SSFT_MODE_DIR:
    case SSFT_MODE_DELETE:
    case SSFT_MODE_MKDIR:
        if (argc != 1)
        {
            log << "Bad number of arguments. Can only have one argument "
                "for this command." << ENDL;
            return -1;   
        }
        break;
    default:
      assert(false);
      return -1;
    }

    int index = 0;
    string dst_path;
    arglist_t src_list;
    bool client_to_server = true;

    if (!extract_host(argc, &argv, dst_path, client_to_server))
        return -1;
    
    for(index = 0; index < argc; index++)
    {
        src_list.push_back(argv[index]);
    }

    if (hostname.empty())
    {
        log << LOG_DATE << "Error: No host is specified" << ENDL;
        return -1;
    }
    
    if (dst_path.empty())
    {
        log << LOG_DATE << "Error: No target path is specified" << ENDL;
        return -1;
    }
    
    if (!resolve_hostname())
        return -1;

    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    SSL_METHOD *meth = SSLv23_client_method();
    SSL_CTX *ctx = SSL_CTX_new(meth);
    if (!ctx) 
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_new() failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -2;
    }

    if (SSL_CTX_load_verify_locations(ctx, rootcert_file.c_str(), NULL) <= 0)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_load_verify_locations(\""
	    << rootcert_file
	    << "\") failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -3;
    }

    if (SSL_CTX_set_default_verify_paths(ctx) <= 0)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_load_verify_locations() failed " 
            << ssl_err << ENDL;        

        SSL_CTX_free (ctx);
        return -4; 
    }

    if (SSL_CTX_use_certificate_file(ctx, cli_cert_file.c_str(), SSL_FILETYPE_PEM) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - SSL_CTX_use_certificate_file() failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -5;
    }
    
    if (SSL_CTX_use_PrivateKey_file(ctx, cli_key_file.c_str(), SSL_FILETYPE_PEM) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - SSL_CTX_use_certificate_file() failed " 
            << ssl_err << ENDL;
        
        return -6;
    }

    if (debug_flag)
        log << "Using client certificate: " << cli_cert_file << ENDL;

    if (SSL_CTX_check_private_key(ctx) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - Private key does not match the certificate public key " 
            << ENDL;
        
        SSL_CTX_free (ctx);
        return -7;
    }
    
    SOCKET server = create_socket(false);
    
    log << "Connecting to "
        << inet_ntoa(pin->sin_addr) 
        << ":" 
        << ntohs(pin->sin_port) 
        << ENDL;
    
    if (connect(server, &socaddr, sizeof(socaddr)) == -1)
    {
        log << LOG_DATE  
            << "failed to connect to server " 
            << GET_SOCKET_ERROR
            << ENDL;

        SSL_CTX_free (ctx);
        return -8;
    }

    ClientConnection my_client(server, ctx);
    if (!my_client.Create())
    {
        SSL_CTX_free(ctx);
        return -9;
    }

    log << LOG_DATE  << "connected to server" << ENDL;

    arglist_t::const_iterator P;

    switch(curr_cmd)
    {
    case SSFT_MODE_COPY:
        for(P = src_list.begin()
            ; P != src_list.end()
            ; ++P)
        {   
            try
            {
                if (client_to_server)
                    my_client.SendFile(P->c_str(), dst_path);
                else // Server to client
                    my_client.RecvFile(P->c_str(), dst_path);
                
            }
            catch(TransferException& e)
            {
                log << LOG_DATE  
                    << e.what()
                    << "Transfer failed" << ENDL;
                
                // Not fatal
            }
            catch(...)
            {
                log << LOG_DATE  
                    << "Transfer failed" << ENDL;
                
                SSL_CTX_free(ctx);
                return -10;
            }
        }
        break;
    case SSFT_MODE_REPLICATE:
        try
        {
            do_replicate(src_list.front(), 
                dst_path, 
                client_to_server,
                my_client);
        }
        catch(std::exception& e)
        {
            log << LOG_DATE  
                << e.what()
                << "Transfer failed" << ENDL;
        }
        catch(...)
        {
            log << LOG_DATE  
                << "Transfer failed" << ENDL;
            
            SSL_CTX_free(ctx);
            return -10;
        }
        break;
    case SSFT_MODE_DIR:
        try
        {
            DirHandlerOut<std::ostream> out(std::cout);
            my_client.ListDir(dst_path.c_str(), out);
        }
        catch(std::exception& e)
        {
            log << LOG_DATE  
                << e.what()
                << "Command failed" << ENDL;
        }
        catch(...)
        {
            log << LOG_DATE  
                << "Command failed" << ENDL;
            
            SSL_CTX_free(ctx);
            return -10;
        }
        break;
    case SSFT_MODE_DELETE:
         try
        {
            my_client.DeletePath(dst_path.c_str());
        }
        catch(std::exception& e)
        {
            log << LOG_DATE  
                << e.what()
                << "Command failed" << ENDL;
        }
        catch(...)
        {
            log << LOG_DATE  
                << "Command failed" << ENDL;
            
            SSL_CTX_free(ctx);
            return -10;
        }
        break;
    case SSFT_MODE_MKDIR:
         try
        {
            my_client.CreateDir(dst_path.c_str());
        }
        catch(std::exception& e)
        {
            log << LOG_DATE  
                << e.what()
                << "Command failed" << ENDL;
        }
        catch(...)
        {
            log << LOG_DATE  
                << "Command failed" << ENDL;
            
            SSL_CTX_free(ctx);
            return -10;
        }
        break;
    default:
        assert(false);
        break;
    }

    try
    {
        my_client.Quit();
    }
    catch(...)
    {
    }

    my_client.Close();
    
    SSL_CTX_free (ctx);

    log << "Successfully done." << ENDL;
    
    return 0;
}

void do_replicate(const std::string& src, 
                  const std::string& dst,
                  bool clientToServer,
                  ClientConnection& client) 
                  throw(TransferException, SsftDisconnect)
{
    // Build two lists of files.
    fileset_t src_set, dst_set;
    fileset_t *plocal, *pserver;
    int flags = 0;

    if (recurse_flag)
        flags |= DIRFLAG_RECURSE;

    if (hash_digest.empty())
        hash_digest = "sha1"; // Default hash digest if not specified

    if (clientToServer)
    {
        plocal = &src_set;
        pserver = &dst_set;
    }
    else
    {
        plocal = &dst_set;
        pserver = &src_set;
    }

    DirList local_dir;

    if ((local_dir.m_Md = EVP_get_digestbyname(hash_digest.c_str())) == NULL)
    {        
        log << LOG_DATE
            << "do_replicate(): Unknown hash-digest name: "
            << hash_digest
            << ENDL;
        
        throw TransferException();
    }

    // Get local files
    const char *src_path = clientToServer ? src.c_str() : dst.c_str();
    local_dir.Open(src_path,
        flags, src_path, hash_digest.c_str());

    local_dir.GetList(*plocal);

    if (debug_flag)
    {
        log << LOG_DATE  
            << "do_replicate(): Fetched " 
            << (*plocal).size()
            << " entries from local storage."
            << ENDL;
    }

    // Get remote files
    DirHandlerList out(*pserver, dst);
    client.ListDir(clientToServer ? dst.c_str() : src.c_str(), out);

    if (debug_flag)
    {
        log << LOG_DATE  
            << "do_replicate(): Fetched " 
            << (*pserver).size()
            << " entries from server storage."
            << ENDL;
    }

    // Compare the lists to decide what to do
    // Start by deciding what to update
    fileset_t::const_iterator P, PP;
    for(P = src_set.begin()
        ; P != src_set.end()
        ; ++P)
    {
        const FileInfo& src_file = *P;

        if ((PP = dst_set.find(src_file)) == dst_set.end())
        {
update_file:

            // Update the file!
            if (debug_flag)
            {
                log << LOG_DATE
                    << "do_replicate("
                    << src_file.mName
                    << "): UPDATE"
                    << ENDL;
            }
            
            if (src_file.mType == TYPE_FILE)
            {
                std::string src_path = merge_paths(src, src_file.mName, '/');
                std::string dst_path = merge_paths(dst, get_pathname(std::string(src_file.mName)), '/');
                
                if (clientToServer)
                {
                    client.SendFile(src_path.c_str(), dst_path.c_str());
                }
                else // Server to client
                {
                    client.RecvFile(src_path.c_str(), dst_path.c_str());
                }
            }
        }
        else
        {
            const FileInfo& dst_file = *PP;

            if (src_file.mHash != dst_file.mHash)
	    {
                goto update_file;
	    }
        }
    }
    
    // Now, get files to delete
    if (delete_flag)
    {
        fileset_t::reverse_iterator R;
        for(R = dst_set.rbegin()
            ; R != dst_set.rend()
            ; ++R)
        {
            const FileInfo& file = *R;
            
            if ((PP = src_set.find(file)) == src_set.end())
            {
                if (debug_flag)
                {
                    log << LOG_DATE
                        << "do_replicate("
                        << file.mName
                        << "): DELETE"
                        << ENDL;
                }
                
                std::string dst_path = merge_paths(dst, 
                    get_pathname(std::string(file.mName)), '/');
                
                client.DeletePath(dst_path.c_str());
            }
        }
    }
}
