/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/


#ifndef SSFT_H
#define SSFT_H

#define PROGRAM "ssft"

#ifdef VERSION
#  undef VERSION
#endif

#define VERSION "0.16"

#ifndef SSH_INCLUDES_H
#   include "ssft_includes.h"
#endif

#ifndef SSFT_PORT
/* We use a port in the experimental range 
 * (RFC 739) that is still unassigned (reserved)
 * in RFC 1700.
 * The port is also in the privlieged range, to
 * make it blocked by default by most firewalls
 * and personal routers.
 */
#   define SSFT_PORT 230
#endif

#ifdef WIN32
#   define IS_SLASH(a) ((a == '/') || (a == '\\'))
#   define SYSENDL "\r\n"
#   define SYSSLASH '\\'
#else
#   define IS_SLASH(a) (a == '/')
#   define SYSENDL '\n'
#   define SYSSLASH '/'
#endif

#ifndef O_BINARY
#   define O_BINARY 0
#endif

#ifndef O_SEQUENTIAL
#   define  O_SEQUENTIAL 0
#endif

#ifdef WIN32
#   define MKDIR_MODE
#else
#   define MKDIR_MODE , 0775
#endif

#ifndef HAVE_STRICMP
#   ifdef HAVE_STRCASECMP
#       define stricmp strcasecmp
#   else
#       error "No stricmp() substitute!"
#   endif
#endif

#ifndef HAVE_STRNICMP
#   ifdef HAVE_STRNCASECMP
#       define strnicmp strncasecmp
#   else
#       error "No strnicmp() substitute!"
#   endif
#endif

#ifndef min
#   define min(a,b) (a < b ? a : b)
#endif

#ifndef max
#   define max(a,b) (a > b ? a : b)
#endif

#define SSFT_SVR_HELO "HELO SSFT protocol\r\n"
#define SSFT_MAX_HEADER_LEN 1024

#define SSFT_NET_BUFLEN (1024 * 16)
#define SSFT_MAX_PATHLEN 512

// Commands
enum SsftClientModeE
{
    SSFT_MODE_COPY,
    SSFT_MODE_DIR,
    SSFT_MODE_REPLICATE,
    SSFT_MODE_DELETE,
    SSFT_MODE_MKDIR,
    SSFT_MODE_INVALID
};

enum SsftCmdE
{
    SSFT_CMD_QUIT = 0,
    SSFT_CMD_TRANSF_CLI_TO_SEVER = 1,
    SSFT_CMD_TRANSF_SEVER_TO_CLI = 2,
    SSFT_CMD_DIR_SEVER_TO_CLI = 3,
    SSFT_CMD_DELETE_FILE_OR_DIR = 4,
    SSFT_CMD_CREATE_DIR = 5
};

enum SsftResultCodesE
{
    SSFT_RESULT_OPERATION_OK = 0,
    SSFT_RESULT_ACCESS_DENIED = 1,
    SSFT_RESULT_UNSPECIFIED_ERROR = 2,
    SSFT_RESULT_BAD_PROTOCOL = 3,
    SSFT_RESULT_UNKNOWN_COMMAND = 4,
    SSFT_RESULT_HEADER_TOO_LONG = 5,
    SSFT_RESULT_INVALID_BLEN = 6,
    SSFT_RESULT_NETWORK_ERROR = 7,
    SSFT_RESULT_HANDSHAKE_FAILED = 8,
    SSFT_RESULT_FAILED_TO_OPEN_FILE = 9,
    SSFT_RESULT_WRITE_ERROR = 10,
    SSFT_RESULT_READ_ERROR = 11,
    SSFT_RESULT_FILE_TOO_LONG = 12,
    SSFT_RESULT_NO_CERT_RECEIVED = 13,
    SSFT_RESULT_INVALID_REPLY = 14,
    SSFT_RESULT_INVALID_HEADER = 15,
    SSFT_RESULT_NOT_A_DIRECTORY = 16,
    SSFT_RESULT_UNKNOWN_HASH_DIGEST = 17,
    SSFT_RESULT_PATH_NOT_FOUND = 18,
    SSFT_RESULT_SHARING_ERROR = 19
};

enum SsftProtocolVersionsE
{
    SSFT_VERSION_1 = 1
};

enum SectionsE
{
    SECT_GENERAL,
    SECT_USER,
    SECT_INVALID
};

enum TokenTypeE
{
    TOKEN_STRING,
    TOKEN_INT,
    TOKEN_OCT,
    TOKEN_INVALID
};

enum DirNodeTypesE
{
    TYPE_FILE,
    TYPE_DIR
};

// Dir-listing flags
#define DIRFLAG_RECURSE 0x00000001

class ClientConnection;
class SsftDisconnect;
class TransferException;
class SsftWouldBlock;

struct ConfToken
{
    TokenTypeE mType;
    size_t mLength;
    const char *mpName;
    void *mpPtr;
};

std::string merge_paths(const std::string& v1,
                        const std::string& v2,
                        char slash_ch = SYSSLASH);

#include "XmitBuffer.h"
#include "User.h"
#include "DirList.h"
#include "DirHandler.h"

void usage();
int do_client_xmit(int argc, char **argv);
int do_server();
char *get_time();
int do_process(int argc, char **argv);
void do_replicate(const std::string& src, 
                  const std::string& dst,
                  bool clientToServer,
                  ClientConnection& client) 
                  throw(TransferException, SsftDisconnect);
void show_gpl();
SOCKET create_socket(bool doBind);
int set_async_mode(SOCKET sck, bool makeAsync);
bool is_dir(const char *path);
bool is_dir(const std::string& path);
bool is_file(const char *path);
bool is_file(const std::string& path);
std::string resolve_path(const std::string& name);
ssft_flen_t htonll(ssft_flen_t v);
#define ntohll htonll
int verify_callback(int preverify_ok, X509_STORE_CTX *ctx);
int do_certs();
std::string get_pathname(const std::string& filePath);
std::string with_slash(const std::string& v, 
                       char slash_ch = SYSSLASH);
bool config_reload();
void clear_user_list();
bool xstricmp(const char *pattern, const char *str);
bool xstrcmp(const char *pattern, const char *str);
bool set_ownership(const std::string& path, 
                   int unix_mode,
                   const char *user, 
                   const char *group);
bool make_dirs(const std::string& path, int unix_mode,
               const char *user, const char *group);
std::string parse_path(const std::string& path);
bool resolve_hostname();
bool is_device(const char *name);

#ifdef WIN32
int get_registry_value(HKEY root, const char *key, const char *value, char *return_buf, DWORD *return_buf_len);
void get_registry_value(HKEY root, const char *key, const char *value, std::string& return_value, const char *def_value);
void UnixTimeToFileTime(const time_t t, FILETIME& ft);
void FileTimeToUnixTime(const FILETIME& ft, time_t& t);
void FileTimeToSsftTime(const FILETIME& ft, ssft_date_t& t);
#endif

extern int my_port;
extern long my_host;
extern int verbose_flag;
extern int debug_flag;
extern int recurse_flag;
extern int nocert_flag;
extern int server_mode;
extern int daemon_mode;
extern int alarm_delay;
extern int delete_flag;
extern char *myname;
extern std::string hostname;
extern std::string logfile_name;
extern int with_passwd_flag;
extern int create_only_flag;
extern int cert_new_ca, cert_new_cli, cert_new_svr;
extern std::string cert_file;
extern std::string key_file;
extern std::string cli_cert_file;
extern std::string cli_key_file;
extern std::string rootcert_file;
extern std::string cert_dir;
extern std::string config_file;
extern std::string pid_file;
extern std::string hash_digest;
extern std::string date_fmt;
extern std::string sign_cert;
extern bool do_continue;
extern struct sockaddr socaddr;
extern struct sockaddr_in *pin;
extern int sockaddr_len;
extern userptr_list_t user_list;
extern SsftClientModeE curr_cmd;

class SslError
{
public:
    SslError()
    {
        char buf[1024 * 4];
        mErrno = (int)ERR_get_error();
#ifdef HAVE_ERR_ERROR_STRING_N
        ERR_error_string_n(mErrno, buf, sizeof(buf) -1);
#else
        ERR_error_string(mErrno, buf);
#endif
        ERR_clear_error();
        mErrorString = buf;
    }

    SslError(SSL *s,int ret_code)
    {
        char buf[1024 * 4];
        mErrno = (int)SSL_get_error(s, ret_code);
#ifdef HAVE_ERR_ERROR_STRING_N
        ERR_error_string_n(mErrno, buf, sizeof(buf) -1);
#else
        ERR_error_string(mErrno, buf);
#endif
        ERR_clear_error();
        mErrorString = buf;
    }

    const std::string& GetErrorString() const
    {
        return mErrorString;
    }

    const int GetErrorCode() const
    {
        return mErrno;
    };

private:
    std::string mErrorString;
    int mErrno;
};

class TransferException : public std::exception
{
public:
    TransferException(const char *why = NULL, 
        const SsftResultCodesE resultCode = SSFT_RESULT_UNSPECIFIED_ERROR)
        : mWhy(why ? why : "(NULL)"),
      mResultCode(resultCode)
    {
    }

    TransferException(const TransferException& from)
    {
      operator = (from);
    }

    TransferException& operator = (const TransferException& from)
    {
      mWhy = from.mWhy;
      mResultCode = from.mResultCode;

      return *this;
    }

    virtual const char *what() const throw()
    {
        return mWhy.c_str();
    }

    const SsftResultCodesE GetResult() const
    {
        return mResultCode;
    }
    
private:
    std::string mWhy;
    SsftResultCodesE mResultCode;
};

class ProgramLog 
{
public:

    ProgramLog();
    ~ProgramLog();

    bool mDoConsoleLog;
};

enum LogE
{
    LOG_DATE,
    ENDL
};

ProgramLog& operator << (ProgramLog& o, const LogE& v);
ProgramLog& operator << (ProgramLog& o, const SslError& v);
ProgramLog& operator << (ProgramLog& o, const sockaddr_in& v);
ProgramLog& operator << (ProgramLog& o, const SsftResultCodesE& v);
ProgramLog& operator << (ProgramLog& o, const SsftCmdE& v);

#ifdef WIN32
ProgramLog& operator << (ProgramLog& o, const __int64& v);
#endif // WIN32

extern ProgramLog log;

template <class T>
ProgramLog& operator << (ProgramLog& o, const T& v)
{
    if (o.mDoConsoleLog)
        std::cout << v;

    if (log_file)
        log_file << v;

    return o;
}


class ClientConnection
{
public:
    enum StateE
    {
        STATE_CREATED,
        STATE_CONNECTED,
        STATE_SENT_FIRST_HEADER,
        STATE_SENT_CMD_HEADER,
        STATE_SENDING,
        STATE_CLOSED
    };

    ClientConnection(SOCKET conn, SSL_CTX *ctx);
    ~ClientConnection();
    bool Create() throw(SsftDisconnect);
    void ListDir(const char *filename, 
        DirHandler& outHandler)
        throw(SsftDisconnect, TransferException);
    void SendFile(const char *filename, 
        const std::string& dst_path) 
         throw(SsftDisconnect, TransferException);
    void RecvFile(const char *filename, 
        const std::string& dst_path) 
         throw(SsftDisconnect, TransferException);
    SsftResultCodesE DeletePath(const char *filename) 
        throw(SsftDisconnect);
    SsftResultCodesE CreateDir(const char *filename) 
        throw(SsftDisconnect);
    void Close();
    void Quit();

    inline SOCKET GetSocket()
    {
        return mSocket;
    }

    inline StateE GetState() const
    {
        return mState;
    }

    const char *GetStateName() const;
    std::string Explain() const;
    void SetState(StateE newState);
    std::vector<char> mOutBuf;
    
private:
    bool Send(const std::vector<char>& sndBuf) throw(SsftDisconnect);
    bool Send() throw(SsftDisconnect);
    SsftResultCodesE GetServerReply() throw(SsftDisconnect);
    void GetServerMessage() throw(SsftDisconnect);
    XmitBuffer mReplyHdr;
    SsftResultCodesE GenericCmd(const SsftCmdE cmd,
        const char *filepath) throw(SsftDisconnect);

    struct sockaddr_in client_addr;
    SOCKET mSocket;
    StateE mState;
    SSL_CTX *mpCtx;
    SSL *mpSsl;

    SsftProtocolVersionsE mServerVersion;
    SsftCmdE mCurrentCommand;
};

class SvrConnection
{
public:
    enum StateE
    {
        STATE_CREATED,
        STATE_CONNECTED,
        STATE_WANT_COMMAND,
        STATE_GOT_CMD_HEADER_LEN,
        STATE_GOT_CMD_HEADER,
        STATE_RECEIVING,
        STATE_SENDING
    };

    SvrConnection(SOCKET conn, 
        SSL_CTX *ctx,
        const sockaddr_in *pCliAddr);
    ~SvrConnection();
    bool Create();

    inline SOCKET GetSocket()
    {
        return mSocket;
    }

    inline StateE GetState() const
    {
        return mState;
    }

    inline bool WantToSend() const
    {
        return mOutputBuffer.empty() == false;
    }

    const char *GetStateName() const;
    const char *GetCmdName() const;
    std::string Explain() const;
    void SetState(StateE newState);
    
    // Callbacks

    void OnRead() throw(SsftDisconnect); // Ready to read data
    void OnWrite() throw(SsftWouldBlock, SsftDisconnect); // Ready to send data

private:
    void WriteResult(const SsftResultCodesE result);
    bool Read() throw(SsftDisconnect);
    int SslAccept();
    bool Write(const char *buffer, size_t length);
    void CloseFile(int error);
    bool PrcName(bool doCreateDirs = true)
        throw(SsftDisconnect);
    SsftResultCodesE DeletePath(int flags,
        const char *path = NULL);
    SsftResultCodesE CreateDir(int flags);
    void OnTransferOk();

    struct sockaddr_in client_addr;
    SOCKET mSocket;
    StateE mState;
    SSL_CTX *mpCtx;
    SSL *mpSsl;
    X509 *mpClientCert;
    User mUser;

    // Input data
    size_t mPendingBytesIn; // Header lenght
    size_t mCurrentBytesReceived;
    XmitBuffer mInputBuffer;

    // Output data
    std::vector<char> mOutputBuffer;

    // Directory listing
    DirList mDir;

    // Client information
    SsftProtocolVersionsE mClientVersion;
    SsftCmdE mCurrentCommand;
    std::string mCurrentFileName; // from client
    std::string mCurrentDstPath; // from client
    std::string mCurrentFilePath; // The actual name
    ssft_flen_t mCurrentFileLength;
    ssft_date_t mCurrentFileDate;
    int mFile;
    ssft_flen_t mCurrentBytesXmitted;
    bool mIsInTransfer;
};

class SsftWouldBlock : 
public std::exception
{
public:
    SsftWouldBlock()
    {}
};

class SsftDisconnect : 
public std::exception
{
public:
    SsftDisconnect(int reason = 0, 
        const char *explain = NULL)
        : mReason(reason),
        mExplanation(explain ? explain: "")
    {
    };

    SsftDisconnect(const SsftDisconnect& from)
        : mReason(from.mReason), 
        mExplanation(from.mExplanation)
    {}

    SsftDisconnect& operator = (const SsftDisconnect& from)
    {
        mReason = from.mReason;
        mExplanation = from.mExplanation;
        return *this;
    }


    virtual const char *what() const
    {
        return mExplanation.empty() 
            ? "No explanation"
            : mExplanation.c_str();
    }

private:
    int mReason;
    std::string mExplanation;
};

typedef std::list<SvrConnection *> ssft_svr_conn_list_t;


#endif SSFT_H

